# ================================================================================ #
#   Authors: Fabio Frazao and Oliver Kirsebom                                      #
#   Contact: fsfrazao@dal.ca, oliver.kirsebom@dal.ca                               #
#   Organization: MERIDIAN (https://meridian.cs.dal.ca/)                           #
#   Team: Data Analytics                                                           #
#   Project: ketos                                                                 #
#   Project goal: The ketos library provides functionalities for handling          #
#   and processing acoustic data and applying deep neural networks to sound        #
#   detection and classification tasks.                                            #
#                                                                                  #
#   License: GNU GPLv3                                                             #
#                                                                                  #
#       This program is free software: you can redistribute it and/or modify       #
#       it under the terms of the GNU General Public License as published by       #
#       the Free Software Foundation, either version 3 of the License, or          #
#       (at your option) any later version.                                        #
#                                                                                  #
#       This program is distributed in the hope that it will be useful,            #
#       but WITHOUT ANY WARRANTY; without even the implied warranty of             #
#       MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the              #
#       GNU General Public License for more details.                               # 
#                                                                                  #
#       You should have received a copy of the GNU General Public License          #
#       along with this program.  If not, see <https://www.gnu.org/licenses/>.     #
# ================================================================================ #

""" 'audio.audio_loader' module within the ketos library

    This module contains the utilities for loading waveforms and computing spectrograms.

    Contents:
        AudioLoader class:
        AudioSelectionLoader class:
        AudioSequenceLoader class
"""
import os
import copy
import numpy as np
import librosa
import warnings
from ketos.audio.waveform import Waveform
from ketos.audio.gammatone import GammatoneFilterBank,AuralFeatures
from ketos.audio.spectrogram import Spectrogram,MagSpectrogram,PowerSpectrogram,MelSpectrogram,CQTSpectrogram
from ketos.data_handling.data_handling import find_wave_files
from ketos.data_handling.selection_table import query


""" Audio representation dictionary 
"""
audio_repres_dict = {'Waveform':Waveform,
                     'MagSpectrogram':MagSpectrogram, 
                     'Mag':MagSpectrogram,
                     'PowerSpectrogram':PowerSpectrogram,
                     'Power':PowerSpectrogram,
                     'Pow':PowerSpectrogram,
                     'MelSpectrogram':MelSpectrogram,
                     'Mel':MelSpectrogram,
                     'CQTSpectrogram':CQTSpectrogram,
                     'CQT':CQTSpectrogram,
                     'AuralFeatures': AuralFeatures,
                     'Aural': AuralFeatures,
                     'GammatoneFilterBank': GammatoneFilterBank,
                     'Gammatone': GammatoneFilterBank}


class SelectionGenerator():
    """ Template class for selection generators.
    """
    def __iter__(self):
        return self

    def __next__(self):
        """ Returns the next audio selection.

            An audio selection is represented as a dictionary 
            with two required keys (data_dir, filename) and 
            an unlimited number of optional keys, which typically 
            include offset, duration, and label.
        
            Must be implemented in child class.

            Returns:
                : dict()
                    Next audio selection
        """
        pass

    def num(self):
        """ Returns total number of selections.
        
            Must be implemented in child class.

            Returns:
                : int
                    Total number of selections.
        """
        pass

    def reset(self):
        """ Resets the selection generator to the beginning.
        """        
        pass
    
class SelectionTableIterator(SelectionGenerator):
    """ Iterates over entries in a selection table.

        Args: 
            data_dir: str
                Path to top folder containing audio files.
            selection_table: pandas DataFrame
                Selection table
            duration: float
                Use this argument to enforce uniform duration of all selections.
                Any selection longer than the specified duration will be shortened
            include_attrs: bool
                If True, load data from all attribute columns in the selection table. Default is False.
            attrs: list(str)
                Specify the names of the attribute columns that you wish to load data from. 
                Overwrites include_attrs if specified. If None, all columns will be loaded provided that 
                include_attrs=True.
    """
    def __init__(self, data_dir, selection_table, duration=None, include_attrs=False, attrs=None):
        self.sel = selection_table
        self.duration = duration
        self.dir = data_dir
        self.row_id = 0

        all_attrs = list(self.sel.columns.values)
        for col in ['start', 'end', 'label']: 
            if col in all_attrs: all_attrs.remove(col)

        if attrs is not None:
            for col in attrs: 
                if col not in all_attrs: attrs.remove(col)
            self.attrs = attrs
        elif include_attrs:
            self.attrs = all_attrs
        else:
            self.attrs = []

    def __next__(self):
        """ Returns the next audio selection.

            Returns:
                audio_sel: dict
                    Audio selection
        """
        audio_sel = self.get_selection(id=self.row_id)
        self.row_id = (self.row_id + 1) % len(self.sel) #update row no.
        return audio_sel

    def num(self):
        """ Returns total number of selections.
        
            Returns:
                : int
                    Total number of selections.
        """
        return len(self.sel)

    def reset(self):
        """ Resets the selection generator to the beginning of the selection table.
        """        
        self.row_id = 0
        
    def get_selection(self, id):
        """ Returns the audio selection with a given id.

            Args:
                id: int
                    The id within the selection table to be searched        
            Returns:
                audio_sel: dict
                    Audio selection
        """
        audio_sel = {'data_dir': self.dir}
        audio_sel['filename'] = self.sel.index.values[id][0]
        # current row
        s = self.sel.iloc[id]
        # start time
        if 'start' in s.keys(): offset = s['start']
        else: offset = 0
        audio_sel['offset'] = offset
        # duration
        if self.duration is not None: audio_sel['duration'] = self.duration
        elif 'end' in s.keys(): audio_sel['duration'] = s['end'] - offset
        # label
        if 'label' in self.sel.columns.values: audio_sel['label'] = s['label']
        # attribute columns
        for col in self.attrs: audio_sel[col] = s[col]
        return audio_sel


class FrameStepper(SelectionGenerator):
    """ Generates selections with uniform duration 'frame', with successive selections 
        displaced by a fixed amount 'step' (If 'step' is not specified, it is set equal 
        to 'frame'.)

        Args: 
            frame: float
                Frame length in seconds.
            step: float
                Separation between consecutive frames in seconds. If None, the step size 
                equals the frame length.
            path: str
                Path to folder containing .wav files. If None is specified, the current directory will be used.
            filename: str or list(str)
                Relative path to a single .wav file or a list of .wav files. Optional.
    """
    def __init__(self, frame, step=None, path=None, filename=None):
        self.frame = frame
        if step is None: self.step = frame
        else: self.step = step

        if path is None: path = os.getcwd()

        # get all wav files in the folder, including subfolders
        if filename is None:
            self.dir = path
            self.files = find_wave_files(path=path, return_path=True, search_subdirs=True)
            assert len(self.files) > 0, '{0} did not find any wave files in {1}'.format(self.__class__.__name__, path)

        else:
            if isinstance(filename, str):
                fullpath = os.path.join(path,filename)
                assert os.path.exists(fullpath), '{0} could not find {1}'.format(self.__class__.__name__, fullpath)
                self.dir = os.path.dirname(fullpath)
                self.files = [os.path.basename(fullpath)]
            else:                
                assert isinstance(filename, list), 'filename must be str or list(str)'        
                self.dir = path
                self.files = filename

        # get file durations
        self.file_durations = np.array([librosa.get_duration(filename=os.path.join(self.dir, f)) for f in self.files])

        # discard any files with 0 second duration
        self.files = np.array(self.files)[self.file_durations > 0].tolist()
        self.file_durations = self.file_durations[self.file_durations > 0].tolist()

        # obtain file durations and compute number of frames for each file
        self.num_segs = [int(np.ceil((dur - self.frame) / self.step)) + 1 for dur in self.file_durations]
        self.num_segs_tot = np.sum(np.array(self.num_segs))

        self.reset()

    def __next__(self):
        """ Returns the next audio selection.
        
            Returns:
                audio_sel: dict
                    Audio selection
        """
        audio_sel = {'data_dir':self.dir, 'filename': self.files[self.file_id], 'offset':self.time, 'duration':self.frame}
        self.time += self.step #increment time       
        self.seg_id += 1 #increment segment ID
        if self.seg_id == self.num_segs[self.file_id]: self._next_file() #if this was the last segment, jump to the next file
        return audio_sel

    def num(self):
        """ Returns total number of selections.
        
            Returns:
                : int
                    Total number of selections.
        """
        return self.num_segs_tot

    def _next_file(self):
        """ Jump to next file. 
        """
        self.file_id = (self.file_id + 1) % len(self.files) #increment file ID
        self.seg_id = 0 #reset
        self.time = 0 #reset

    def reset(self):
        """ Resets the selection generator to the beginning of the first file.
        """        
        self.file_id = -1
        self._next_file()

    def get_file_paths(self, fullpath=True):
        """ Get the paths to the audio files associated with this instance.

            Args:
                fullpath: bool
                    Whether to return the full path (default) or only the filename.

            Returns:
                ans: list
                    List of file paths
        """
        if fullpath:
            ans = [os.path.join(self.dir, f) for f in self.files]
        else:
            ans = self.files

        return ans

    def get_file_durations(self):
        """ Get the durations of the audio files associated with this instance.

            Returns:
                ans: list
                    List of file durations in seconds
        """
        return self.file_durations

class AudioLoader():
    """ Class for loading segments of audio data from .wav files. 

        Several representations of the audio data are possible, including 
        waveform, magnitude spectrogram, power spectrogram, mel spectrogram, 
        and CQT spectrogram.

        Args:
            selection_gen: SelectionGenerator
                Selection generator
            channel: int
                For stereo recordings, this can be used to select which channel to read from
            annotations: pandas DataFrame
                Annotation table
            repres: dict
                Audio data representation. Must contain the key 'type' as well as any arguments 
                required to initialize the class using the from_wav method.  
                
                    * Waveform: 
                        (rate), (resample_method)
                    
                    * MagSpectrogram, PowerSpectrogram, MelSpectrogram: 
                        window, step, (window_func), (rate), (resample_method)
                    
                    * CQTSpectrogram:
                        step, bins_per_oct, (freq_min), (freq_max), (window_func), (rate), (resample_method)

                Optionally, may also contain the key 'normalize_wav' which can have value True or False. 
                If True, the waveform is normalized zero mean (mean=0) and (std=1) unity standard deviation.
                It is also possible to specify multiple audio presentations as a list.
        Examples:
            See child classes :class:`audio.audio_loader.AudioFrameLoader` and 
            :class:`audio.audio_loader.AudioSelectionLoader`.            
    """
    def __init__(self, selection_gen, channel=0, annotations=None, repres={'type': 'Waveform'}, **kwargs):
        repres = copy.deepcopy(repres)
        if not isinstance(repres, list): repres = [repres]
        self.typ, self.cfg = [], []
        for r in repres:
            self.typ.append(r.pop('type'))
            if 'duration' in r.keys(): r.pop('duration')
            self.cfg.append(r)

        self.channel = channel
        self.sel_gen = selection_gen
        self.annot = annotations
        self.kwargs = kwargs

    def __iter__(self):
        return self

    def __next__(self):
        """ Load next waveform segment or compute next spectrogram.

            Returns: 
                : Waveform or Spectrogram
                    Next segment
        """
        audio_sel = next(self.sel_gen)
        return self.load(**audio_sel, **self.kwargs)

    def num(self):
        """ Returns total number of segments.
        
            Returns:
                : int
                    Total number of segments.
        """
        return self.sel_gen.num()

    def load(self, data_dir, filename, offset=0, duration=None, label=None, apply_transforms=True, **kwargs):
        """ Load audio segment for specified file and time.

            Args:
                data_dir: str
                    Data directory
                filename: str
                    Filename or relative path
                offset: float
                    Start time of the segment in seconds, measured from the 
                    beginning of the file.
                duration: float
                    Duration of segment in seconds.
                label: int
                    Integer label
                apply_transforms: bool
                    Apply transforms. Default is True.
        
            Returns: 
                seg: BaseAudio
                    Audio segment
        """
        path = os.path.join(data_dir, filename)

        # load audio
        # (ignore warnings from the from_wav method)
        segs = []
        for typ,cfg in zip(self.typ, self.cfg):

            with warnings.catch_warnings():
                warnings.simplefilter("ignore")        
                _kwargs = kwargs.copy()
                _kwargs.update(cfg)
                if not apply_transforms and 'transforms' in _kwargs.keys(): del _kwargs['transforms']
                seg = audio_repres_dict[typ].from_wav(path=path, channel=self.channel, offset=offset, 
                                                            duration=duration, id=filename, **_kwargs)
        
            # add annotations
            if label is not None:
                seg.label = label

            if self.annot is not None:
                q = query(self.annot, filename=os.path.basename(path), start=offset, end=offset+duration)
                if len(q) > 0:
                    q['start'] = np.maximum(0, q['start'].values - offset)
                    q['end']   = np.minimum(q['end'].values - offset, seg.duration())
                    seg.annotate(df=q)  

            segs.append(seg)           

        if len(segs)==1: segs=segs[0]

        return segs

    def reset(self):
        """ Resets the audio loader to the beginning.
        """        
        self.sel_gen.reset()

class AudioFrameLoader(AudioLoader):
    """ Load segments of audio data from .wav files. 

        Loads segments of uniform duration 'frame', with successive segments
        displaced by an amount 'step'. (If 'step' is not specified, it is 
        set equal to 'frame'.)

        Args:
            frame: float
                Segment duration in seconds. Can also be specified via the 'duration' 
                item of the 'repres' dictionary.
            step: float
                Separation between consecutive segments in seconds. If None, the step size 
                equals the segment duration.
            path: str
                Path to folder containing .wav files. If None is specified, the current directory will be used.
            filename: str or list(str)
                relative path to a single .wav file or a list of .wav files. Optional
            channel: int
                For stereo recordings, this can be used to select which channel to read from
            annotations: pandas DataFrame
                Annotation table
            repres: dict
                Audio data representation. Must contain the key 'type' as well as any arguments 
                required to initialize the class using the from_wav method.  
                It is also possible to specify multiple audio presentations as a list. These 
                presentations must have the same duration.
            batch_size: int or str
                Load segments in batches rather than one at the time. 
                Increasing the batch size can help reduce computational time.
                The default batch size is 1. 
                You can also specify batch_size='file' to load one wav file at the time.

        Examples:
            >>> import librosa
            >>> from ketos.audio.audio_loader import AudioFrameLoader
            >>> # specify path to wav file
            >>> filename = 'ketos/tests/assets/2min.wav'
            >>> # check the duration of the audio file
            >>> print(librosa.get_duration(filename=filename))
            120.832
            >>> # specify the audio representation
            >>> rep = {'type':'MagSpectrogram', 'window':0.2, 'step':0.02, 'window_func':'hamming', 'freq_max':1000.}
            >>> # create an object for loading 30-s long spectrogram segments, using a step size of 15 s (50% overlap) 
            >>> loader = AudioFrameLoader(frame=30., step=15., filename=filename, repres=rep)
            >>> # print number of segments
            >>> print(loader.num())
            8
            >>> # load and plot the first segment
            >>> spec = next(loader)
            >>>
            >>> import matplotlib.pyplot as plt
            >>> fig = spec.plot()
            >>> fig.savefig("ketos/tests/assets/tmp/spec_2min_0.png")
            >>> plt.close(fig)
            
            .. image:: ../../../ketos/tests/assets/tmp/spec_2min_0.png
    """
    def __init__(self, frame=None, step=None, path=None, filename=None, channel=0, 
                    annotations=None, repres={'type': 'Waveform'}, batch_size=1, **kwargs):

        if isinstance(repres, list): r0 = repres[0]
        else: r0 = repres

        assert 'duration' in r0.keys() or frame is not None, 'duration must be specified either via the frame argument or the duration item of the repres dictionary'

        if frame is None: frame = r0['duration']

        if 'duration' in r0.keys() and r0['duration'] is not None and r0['duration'] != frame:
            print("Warning: Mismatch between frame size ({0:.3f} s) and duration ({1:.3f} s). The latter value will be ignored.")

        assert (isinstance(batch_size, int) and batch_size >= 1) or (isinstance(batch_size, str) and batch_size.lower() == 'file'), 'Batch size must be a positive integer or have the string value file'

        super().__init__(selection_gen=FrameStepper(frame=frame, step=step, path=path, filename=filename), 
            channel=channel, annotations=annotations, repres=repres, **kwargs)

        self.transforms_list = []
        for config in self.cfg:
            transforms = config['transforms'] if 'transforms' in config.keys() else []
            self.transforms_list.append(transforms)

        if isinstance(batch_size, int):
            self.max_batch_size = batch_size
        else:
            self.max_batch_size = np.inf

        if self.max_batch_size > 1:
            audio_sel = next(self.sel_gen)
            self.offset = audio_sel['offset']
            self.data_dir = audio_sel['data_dir']
            self.filename = audio_sel['filename']
            self.load_next_batch()

    def __next__(self):
        """ Load next waveform segment or compute next spectrogram.

            Returns: 
                : Waveform or Spectrogram
                    Next segment
        """
        if self.max_batch_size == 1:
            return super().__next__()        
        else:
            return self.next_in_batch()

    def load_next_batch(self):
        """ Load the next batch of waveforms or spectrograms.
        """
        self.batch_size = 0
        self.counter = 0
        offset = np.inf
        data_dir = self.data_dir
        filename = self.filename
        while data_dir == self.data_dir and filename == self.filename and offset > self.offset and self.batch_size < self.max_batch_size:
            self.batch_size += 1
            audio_sel = next(self.sel_gen)
            offset = audio_sel['offset']
            data_dir = audio_sel['data_dir']
            filename = audio_sel['filename']            

        duration = self.sel_gen.frame + self.sel_gen.step * (self.batch_size - 1)

        # load the data without applying transforms
        self.batch = self.load(data_dir=self.data_dir, filename=self.filename, offset=self.offset, 
            duration=duration, label=None, apply_transforms=False, **self.kwargs)

        if not isinstance(self.batch, list): self.batch = [self.batch]

        # loop over the representations
        for i in range(len(self.transforms_list)):

            # segment the data
            self.batch[i] = self.batch[i].segment(window=self.sel_gen.frame, step=self.sel_gen.step)

            # apply the transforms to each segment separately 
            with warnings.catch_warnings():
                warnings.simplefilter("ignore")        
                for j in range(len(self.batch[i])):
                    self.batch[i][j].apply_transforms(self.transforms_list[i])


        if len(self.batch) == 1: self.batch = self.batch[0]

        self.offset = offset
        self.data_dir = data_dir
        self.filename = filename

    def next_in_batch(self):
        """ Load the next waveform or spectrogram in the batch.
        
            Returns: 
                a: Waveform or Spectrogram
                    Next segment
        """
        if self.counter >= len(self.batch): self.load_next_batch()
        a = self.batch[self.counter]
        self.counter += 1
        return a

    def get_file_paths(self, fullpath=True):
        """ Get the paths to the audio files associated with this instance.

            Args:
                fullpath: bool
                    Whether to return the full path (default) or only the filename.

            Returns:
                ans: list
                    List of file paths
        """
        return self.sel_gen.get_file_paths(fullpath=fullpath)

    def get_file_durations(self):
        """ Get the durations of the audio files associated with this instance.

            Returns:
                ans: list
                    List of file durations in seconds
        """
        return self.sel_gen.get_file_durations()

class AudioSelectionLoader(AudioLoader):
    """ Load segments of audio data from .wav files. 

        The segments to be loaded are specified via a selection table.

        Args:
            selections: pandas DataFrame
                Selection table
            path: str
                Path to folder containing .wav files
            filename: str or list(str)
                relative path to a single .wav file or a list of .wav files. Optional
            annotations: pandas DataFrame
                Annotation table
            repres: dict
                Audio data representation. Must contain the key 'type' as well as any arguments 
                required to initialize the class using the from_wav method.  
                It is also possible to specify multiple audio presentations as a list.
            include_attrs: bool
                If True, load data from all attribute columns in the selection table. Default is False.
            attrs: list(str)
                Specify the names of the attribute columns that you wish to load data from. 
                Overwrites include_attrs if specified. If None, all columns will be loaded provided that 
                include_attrs=True.
    """
    def __init__(self, path, selections, channel=0, annotations=None, repres={'type': 'Waveform'}, 
        include_attrs=False, attrs=None, **kwargs):

        if isinstance(repres, list): r0 = repres[0]
        else: r0 = repres

        if 'duration' in r0.keys(): duration = r0['duration']
        else: duration = None

        super().__init__(selection_gen=SelectionTableIterator(data_dir=path, selection_table=selections, duration=duration, include_attrs=include_attrs, attrs=attrs), 
            channel=channel, annotations=annotations, repres=repres, **kwargs)


# ================================================================================ #
#   Authors: Fabio Frazao and Oliver Kirsebom                                      #
#   Contact: fsfrazao@dal.ca, oliver.kirsebom@dal.ca                               #
#   Organization: MERIDIAN (https://meridian.cs.dal.ca/)                           #
#   Team: Data Analytics                                                           #
#   Project: ketos                                                                 #
#   Project goal: The ketos library provides functionalities for handling          #
#   and processing acoustic data and applying deep neural networks to sound        #
#   detection and classification tasks.                                            #
#                                                                                  #
#   License: GNU GPLv3                                                             #
#                                                                                  #
#       This program is free software: you can redistribute it and/or modify       #
#       it under the terms of the GNU General Public License as published by       #
#       the Free Software Foundation, either version 3 of the License, or          #
#       (at your option) any later version.                                        #
#                                                                                  #
#       This program is distributed in the hope that it will be useful,            #
#       but WITHOUT ANY WARRANTY; without even the implied warranty of             #
#       MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the              #
#       GNU General Public License for more details.                               # 
#                                                                                  #
#       You should have received a copy of the GNU General Public License          #
#       along with this program.  If not, see <https://www.gnu.org/licenses/>.     #
# ================================================================================ #

""" Unit tests for the 'audio.audio_loader' module within the ketos library
"""
import pytest
import json
import os
import numpy as np
import pandas as pd
from io import StringIO
from ketos.audio.waveform import Waveform
from ketos.audio.spectrogram import MagSpectrogram
from ketos.audio.audio_loader import AudioFrameLoader, AudioSelectionLoader
from ketos.data_handling.selection_table import use_multi_indexing, standardize
from ketos.data_handling.data_handling import find_wave_files
from ketos.data_handling.parsing import parse_audio_representation
from ketos.audio.utils.misc import from_decibel

current_dir = os.path.dirname(os.path.realpath(__file__))
path_to_assets = os.path.join(os.path.dirname(current_dir),"assets")

def test_init_audio_frame_loader_with_folder(five_time_stamped_wave_files):
    """ Test that we can initialize an instance of the AudioFrameLoader class from a folder"""
    loader = AudioFrameLoader(path=five_time_stamped_wave_files, frame=0.5)
    assert len(loader.sel_gen.files) == 5

def test_init_audio_frame_loader_with_wav_file(sine_wave_file):
    """ Test that we can initialize an instance of the AudioFrameLoader class 
        from a single wav file"""
    loader = AudioFrameLoader(filename=sine_wave_file, frame=0.5)
    assert len(loader.sel_gen.files) == 1
    assert loader.num() == 6

def test_init_audio_frame_loader_with_batches(sine_wave_file):
    """ Test that we can initialize an instance of the AudioFrameLoader class 
        from a single wav file with a batch size greater than 1"""
    loader = AudioFrameLoader(filename=sine_wave_file, frame=0.5, batch_size=2)
    assert len(loader.sel_gen.files) == 1
    assert loader.num() == 6

def test_init_audio_frame_loader_with_batch_size_one_file(sine_wave_file):
    """ Test that we can initialize an instance of the AudioFrameLoader class 
        from a single wav file with a batch size equal to 1 file"""
    loader = AudioFrameLoader(filename=sine_wave_file, frame=0.5, batch_size='FILE')
    assert len(loader.sel_gen.files) == 1
    assert loader.num() == 6

def test_audio_frame_loader_gives_same_output_with_batches(sine_wave_file):
    """ Test that segments returned by the AudioFrameLoader class are independent of batch size"""
    rep = {'type':'MagSpectrogram','window':0.1,'step':0.02,'freq_max':800}
    fname = os.path.join(path_to_assets, 'grunt1.wav')
    loader1 = AudioFrameLoader(filename=fname, frame=0.4, step=0.12, repres=rep, batch_size=1)
    loader3 = AudioFrameLoader(filename=fname, frame=0.4, step=0.12, repres=rep, batch_size=3)
    loaderf = AudioFrameLoader(filename=fname, frame=0.4, step=0.12, repres=rep, batch_size='file')
    for i in range(loader1.num()):
        x1 = next(loader1)
        x3 = next(loader3)
        xf = next(loaderf)
        dx = x1.data - x3.data
        assert np.mean(np.abs(dx)) < 0.1
        dx = x1.data - xf.data
        assert np.mean(np.abs(dx)) < 0.1

def test_audio_frame_loader_mag(five_time_stamped_wave_files):
    """ Test that we can use the AudioFrameLoader class to compute MagSpectrograms""" 
    rep = {'type':'MagSpectrogram','window':0.1,'step':0.02,'decibel':False}
    loader = AudioFrameLoader(path=five_time_stamped_wave_files, frame=0.5, repres=rep)
    assert len(loader.sel_gen.files) == 5
    assert loader.num() == 5
    s = next(loader)
    assert s.duration() == 0.5
    s = next(loader)
    assert s.duration() == 0.5
    assert loader.sel_gen.file_id == 2
    loader.reset()
    assert loader.sel_gen.file_id == 0

def test_audio_frame_loader_multiple_representations(five_time_stamped_wave_files):
    """ Test that we can use the AudioFrameLoader class to load multiple audio representations""" 
    rep1 = {'type':'Waveform'}
    rep2 = {'type':'MagSpectrogram','window':0.1,'step':0.02}
    loader = AudioFrameLoader(path=five_time_stamped_wave_files, frame=0.5, repres=[rep1, rep2])
    assert len(loader.sel_gen.files) == 5
    assert loader.num() == 5
    s = next(loader)
    assert len(s) == 2
    assert type(s[0]) == Waveform
    assert type(s[1]) == MagSpectrogram
    assert s[0].duration() == 0.5
    s = next(loader)
    assert s[1].duration() == 0.5
    assert loader.sel_gen.file_id == 2
    loader.reset()
    assert loader.sel_gen.file_id == 0

def test_audio_frame_loader_mag_in_batches(five_time_stamped_wave_files):
    """ Test that we can use the AudioFrameLoader class to compute MagSpectrograms 
        in batches""" 
    rep = {'type':'MagSpectrogram','window':0.1,'step':0.02, 'transforms':[]}
    loader = AudioFrameLoader(path=five_time_stamped_wave_files, frame=0.26, repres=rep, batch_size=3)
    assert len(loader.sel_gen.files) == 5
    assert loader.num() == 10
    s = next(loader) 
    assert loader.sel_gen.file_id == 1
    assert s.duration() == 0.26
    assert s.offset == 0
    s = next(loader)
    assert loader.sel_gen.file_id == 1
    assert s.duration() == 0.26
    assert s.offset == 0.26
    s = next(loader)
    assert loader.sel_gen.file_id == 2
    assert s.duration() == 0.26
    assert s.offset == 0

def test_audio_frame_loader_mag_in_batches_1_file(five_time_stamped_wave_files):
    """ Test that we can use the AudioFrameLoader class to compute MagSpectrograms 
        in batches of 1 file per batch""" 
    rep = {'type':'MagSpectrogram','window':0.1,'step':0.02}
    loader = AudioFrameLoader(path=five_time_stamped_wave_files, frame=0.12, repres=rep, batch_size='file')
    assert len(loader.sel_gen.files) == 5
    assert loader.num() == 25
    assert loader.sel_gen.file_id == 1
    s = next(loader) 
    assert loader.sel_gen.file_id == 1
    assert s.duration() == 0.12
    assert s.offset == 0
    s = next(loader)
    assert loader.sel_gen.file_id == 1
    assert s.duration() == 0.12
    assert s.offset == 0.12
    s = next(loader)
    s = next(loader)
    s = next(loader)
    assert loader.sel_gen.file_id == 1
    s = next(loader)
    assert loader.sel_gen.file_id == 2

def test_audio_frame_loader_norm_mag(sine_wave_file):
    """ Test that we can initialize the AudioFrameLoader class to compute MagSpectrograms
        with the normalize_wav option set to True""" 
    rep = {'type':'MagSpectrogram','window':0.1,'step':0.02}
    loader = AudioFrameLoader(filename=sine_wave_file, frame=0.5, repres=rep)
    spec1 = next(loader)
    spec1 = next(loader)
    rep = {'type':'MagSpectrogram','window':0.1,'step':0.02, 'normalize_wav': True}
    loader = AudioFrameLoader(filename=sine_wave_file, frame=0.5, repres=rep)
    spec2 = next(loader)
    spec2 = next(loader)
    d1 = from_decibel(spec1.get_data())
    d2 = from_decibel(spec2.get_data()) / np.sqrt(2)
    assert np.all(np.isclose(np.mean(d1), np.mean(d2), rtol=2e-2))

def test_audio_frame_loader_mag_transforms(sine_wave_file):
    """ Test that we can initialize the AudioFrameLoader class to compute MagSpectrograms
        with various transformations applied""" 
    range_trans = {'name':'adjust_range', 'range':(0,1)}
    enh_trans = {'name':'enhance_signal','enhancement':2.3}
    transforms = [range_trans, enh_trans]
    norm_trans = {'name':'normalize','mean':0.5,'std':2.0}
    noise_trans = {'name':'add_gaussian_noise', 'sigma':2.0}
    wf_transforms = [norm_trans, noise_trans]
    rep = {'type':'MagSpectrogram','window':0.1,'step':0.02, 'transforms':transforms, 'waveform_transforms':wf_transforms}
    loader = AudioFrameLoader(filename=sine_wave_file, frame=0.5, repres=rep)
    spec1 = next(loader)
    assert spec1.transform_log == transforms
    assert spec1.waveform_transform_log == wf_transforms

def test_audio_frame_loader_dur(five_time_stamped_wave_files):
    """ Test that we can use the AudioFrameLoader class to compute MagSpectrograms
        with durations shorter than file durations""" 
    rep = {'type':'MagSpectrogram','window':0.1,'step':0.02}
    loader = AudioFrameLoader(path=five_time_stamped_wave_files, frame=0.2, repres=rep)
    assert len(loader.sel_gen.files) == 5
    s = next(loader)
    assert s.duration() == 0.2
    s = next(loader)
    assert s.duration() == 0.2
    s = next(loader)
    assert s.duration() == 0.2
    assert loader.sel_gen.file_id == 1

def test_audio_frame_loader_frame_None(five_time_stamped_wave_files):
    """ Test that we can use the AudioFrameLoader class to compute MagSpectrograms
        without specifying the frame argument""" 
    rep = {'type':'MagSpectrogram','window':0.1,'step':0.02,'duration':0.2}
    loader = AudioFrameLoader(path=five_time_stamped_wave_files, repres=rep)
    assert len(loader.sel_gen.files) == 5
    s = next(loader)
    assert s.duration() == 0.2
    s = next(loader)
    assert s.duration() == 0.2
    s = next(loader)
    assert s.duration() == 0.2
    assert loader.sel_gen.file_id == 1

def test_audio_frame_loader_overlap(five_time_stamped_wave_files):
    """ Test that we can use the AudioFrameLoader class to compute overlapping 
        MagSpectrograms""" 
    rep = {'type':'MagSpectrogram','window':0.1,'step':0.02}
    loader = AudioFrameLoader(path=five_time_stamped_wave_files, frame=0.2, step=0.06, repres=rep)
    assert len(loader.sel_gen.files) == 5
    s = next(loader)
    assert s.duration() == 0.2
    s = next(loader)
    assert s.duration() == 0.2
    s = next(loader)
    assert s.duration() == 0.2
    assert loader.sel_gen.time == pytest.approx(3*0.06, abs=1e-6)
    assert loader.sel_gen.file_id == 0

def test_audio_frame_loader_overlap_batches(five_time_stamped_wave_files):
    """ Test that we can use the AudioFrameLoader class to compute overlapping 
        MagSpectrograms in batches""" 
    rep = {'type':'MagSpectrogram','window':0.1,'step':0.02}
    loader = AudioFrameLoader(path=five_time_stamped_wave_files, frame=0.2, step=0.06, repres=rep, batch_size=2)
    assert len(loader.sel_gen.files) == 5
    s = next(loader)
    assert s.duration() == 0.2
    s = next(loader)
    assert s.duration() == 0.2
    s = next(loader)
    assert s.duration() == 0.2
    assert loader.sel_gen.time == pytest.approx(5*0.06, abs=1e-6)
    assert loader.sel_gen.file_id == 0

def test_audio_frame_loader_uniform_length(five_time_stamped_wave_files):
    """ Check that the AudioFrameLoader always returns segments of the same length""" 
    rep = {'type':'MagSpectrogram','window':0.1,'step':0.02}
    loader = AudioFrameLoader(path=five_time_stamped_wave_files, frame=0.2, repres=rep)
    assert len(loader.sel_gen.files) == 5
    for _ in range(10):
        s = next(loader)
        assert s.duration() == 0.2

def test_audio_frame_loader_number_of_segments(sine_wave_file):
    """ Check that the AudioFrameLoader computes expected number of segments""" 
    rep = {'type':'MagSpectrogram','window':0.1,'step':0.01,'rate':2341}
    import librosa
    dur = librosa.core.get_duration(filename=sine_wave_file)
    # duration is an integer number of lengths
    l = 0.2
    loader = AudioFrameLoader(filename=sine_wave_file, frame=l, repres=rep)
    assert len(loader.sel_gen.files) == 1
    N = int(dur / l)
    assert N == loader.sel_gen.num_segs[0]
    # duration is *not* an integer number of lengths
    l = 0.21
    loader = AudioFrameLoader(filename=sine_wave_file, frame=l, repres=rep)
    N = int(np.ceil(dur / l))
    assert N == loader.sel_gen.num_segs[0]
    # loop over all segments
    for _ in range(N):
        _ = next(loader)
    # non-zero overlap
    l = 0.21
    o = 0.8*l
    loader = AudioFrameLoader(filename=sine_wave_file, frame=l, step=l-o, repres=rep)
    step = l - o
    N = int(np.ceil((dur-l) / step) + 1)
    assert N == loader.sel_gen.num_segs[0]
    # loop over all segments
    for _ in range(N):
        _ = next(loader)

def test_audio_select_loader_mag(five_time_stamped_wave_files):
    """ Test that we can use the AudioSelectionLoader class to compute MagSpectrograms""" 
    rep = {'type':'MagSpectrogram','window':0.1,'step':0.02}
    # create a selection table
    files = find_wave_files(path=five_time_stamped_wave_files, return_path=False, search_subdirs=True)
    sel = pd.DataFrame({'filename':[files[0],files[1]],'start':[0.10,0.12],'end':[0.46,0.42]})
    sel = use_multi_indexing(sel, 'sel_id')
    # init loader
    loader = AudioSelectionLoader(path=five_time_stamped_wave_files, selections=sel, repres=rep)
    assert loader.num() == 2
    s = next(loader)
    assert s.duration() == pytest.approx(0.36, abs=1e-6)
    s = next(loader)
    assert s.duration() == pytest.approx(0.30, abs=1e-6)

def test_audio_select_loader_with_labels(five_time_stamped_wave_files):
    """ Test that we can use the AudioSelectionLoader class to compute MagSpectrograms with labels""" 
    rep = {'type':'MagSpectrogram','window':0.1,'step':0.02}
    # create a selection table
    files = find_wave_files(path=five_time_stamped_wave_files, return_path=False, search_subdirs=True)
    sel = pd.DataFrame({'filename':[files[0],files[1]],'start':[0.10,0.12],'end':[0.46,0.42],'label':[3,5]})
    sel = use_multi_indexing(sel, 'sel_id')
    # init loader
    loader = AudioSelectionLoader(path=five_time_stamped_wave_files, selections=sel, repres=rep)
    s = next(loader)
    assert s.duration() == pytest.approx(0.36, abs=1e-6)
    assert s.label == 3
    s = next(loader)
    assert s.duration() == pytest.approx(0.30, abs=1e-6)
    assert s.label == 5

def test_audio_select_loader_with_annots(five_time_stamped_wave_files):
    """ Test that we can use the AudioSelectionLoader class to compute MagSpectrograms
        while including annotation data""" 
    rep = {'type':'MagSpectrogram','window':0.1,'step':0.02}
    files = find_wave_files(path=five_time_stamped_wave_files, return_path=False, search_subdirs=True)
    # create a selection table
    sel = pd.DataFrame({'filename':[files[0],files[1]],'start':[0.10,0.12],'end':[0.46,0.42]})
    sel = use_multi_indexing(sel, 'sel_id')
    # create a annotation table
    ann = pd.DataFrame({'filename':[files[0],files[0],files[1]],'label':[3,5,4],'start':[0.05,0.06,0.20],'end':[0.30,0.16,0.60]})
    ann = standardize(ann)
    # init loader
    loader = AudioSelectionLoader(path=five_time_stamped_wave_files, selections=sel, annotations=ann, repres=rep)
    s = next(loader)
    assert s.duration() == pytest.approx(0.36, abs=1e-6)
    d = '''label  start   end  freq_min  freq_max
0      1    0.0  0.20       NaN       NaN
1      3    0.0  0.06       NaN       NaN'''
    ans = pd.read_csv(StringIO(d), delim_whitespace=True, index_col=[0,1])
    res = s.get_annotations()[ans.columns.values]
    pd.testing.assert_frame_equal(ans, res)
    s = next(loader)
    assert s.duration() == pytest.approx(0.30, abs=1e-6)
    d = '''label  start  end  freq_min  freq_max
0      2   0.08  0.3       NaN       NaN'''
    ans = pd.read_csv(StringIO(d), delim_whitespace=True, index_col=[0,1])
    res = s.get_annotations()[ans.columns.values]
    pd.testing.assert_frame_equal(ans, res)

def test_audio_frame_loader_mag_json(five_time_stamped_wave_files, spectr_settings):
    """ Test that we can use the AudioFrameLoader class to compute MagSpectrograms from json settings""" 
    data = json.loads(spectr_settings)
    rep = parse_audio_representation(data['spectrogram'])
    loader = AudioFrameLoader(path=five_time_stamped_wave_files, frame=0.5, repres=rep)
    assert len(loader.sel_gen.files) == 5
    s = next(loader)
    assert s.duration() == 0.5
    s = next(loader)
    assert s.duration() == 0.5
    assert loader.sel_gen.file_id == 2

def test_audio_frame_loader_accepts_filename_list(five_time_stamped_wave_files, spectr_settings):
    """ Test that we can use the AudioFrameLoader class to compute MagSpectrograms from json settings""" 
    data = json.loads(spectr_settings)
    rep = parse_audio_representation(data['spectrogram'])
    filename = ['empty_HMS_12_ 5_ 0__DMY_23_ 2_84.wav', 
                'empty_HMS_12_ 5_ 1__DMY_23_ 2_84.wav',
                'empty_HMS_12_ 5_ 2__DMY_23_ 2_84.wav']
    loader = AudioFrameLoader(path=five_time_stamped_wave_files, filename=filename, frame=0.5, repres=rep)
    assert len(loader.sel_gen.files) == 3
    s = next(loader)
    assert s.duration() == 0.5
    s = next(loader)
    assert s.duration() == 0.5
    assert loader.sel_gen.file_id == 2

def test_audio_select_loader_stores_source_data(five_time_stamped_wave_files):
    """ Test that we can use the AudioSelectionLoader class to compute MagSpectrograms
        and that the spectrograms retain the correct source data (filename, offset) """ 
    rep = {'type':'MagSpectrogram','window':0.1,'step':0.02}
    # create a selection table
    files = find_wave_files(path=five_time_stamped_wave_files, return_path=False, search_subdirs=True)
    filename = [files[0],files[1]]
    start = [0.10,0.12]
    end = [0.46,0.42]
    sel = pd.DataFrame({'filename':filename,'start':start,'end':end})
    sel = use_multi_indexing(sel, 'sel_id')
    # init loader
    loader = AudioSelectionLoader(path=five_time_stamped_wave_files, selections=sel, repres=rep)
    assert loader.num() == 2
    for i in range(6): #loop over each item 3 times
        s = next(loader)
        assert s.offset == start[i%2]
        assert s.filename == filename[i%2]

def test_audio_frame_loader_on_2min_wav():
    rep = {'type':'MagSpectrogram', 'window':0.2, 'step':0.02, 'window_func':'hamming', 'freq_max':600.}
    path = os.path.join(path_to_assets, '2min.wav')
    loader = AudioFrameLoader(filename=path, frame=30., step=15., repres=rep)
    assert loader.num() == 8
    s = next(loader)
    assert s.freq_max() == pytest.approx(600, abs=s.freq_res())

def test_audio_frame_loader_subdirs():
    """Test that loader can load audio files from subdirectories"""
    rep = {'type':'MagSpectrogram', 'window':0.2, 'step':0.02, 'window_func':'hamming', 'freq_max':1000.}
    path = os.path.join(path_to_assets, 'wav_files')
    loader = AudioFrameLoader(path=path, frame=30., step=15., repres=rep)
    assert len(loader.sel_gen.files) == 3
    for _ in range(loader.num()):
        _ = next(loader)

def test_audio_select_loader_uniform_duration(five_time_stamped_wave_files):
    """ Test that we can use the AudioSelectionLoader class to compute MagSpectrograms
        with uniform duration by specifying duration in audio representation dictionary """ 
    rep = {'type':'MagSpectrogram','window':0.1,'step':0.02,'duration':0.3}
    # create a selection table
    files = find_wave_files(path=five_time_stamped_wave_files, return_path=False, search_subdirs=True)
    sel = pd.DataFrame({'filename':files})
    sel = use_multi_indexing(sel, 'sel_id')
    # init loader
    loader = AudioSelectionLoader(path=five_time_stamped_wave_files, selections=sel, repres=rep)
    assert loader.num() == 5
    for i in range(5):
        s = next(loader)
        assert s.duration() == rep['duration']

def test_audio_select_loader_entire_files(five_time_stamped_wave_files):
    """ Test that we can use the AudioSelectionLoader class to compute MagSpectrograms
        of entire wav files """ 
    rep = {'type':'MagSpectrogram','window':0.1,'step':0.02}
    # create a selection table
    files = find_wave_files(path=five_time_stamped_wave_files, return_path=False, search_subdirs=True)
    sel = pd.DataFrame({'filename':files})
    sel = use_multi_indexing(sel, 'sel_id')
    # init loader
    loader = AudioSelectionLoader(path=five_time_stamped_wave_files, selections=sel, repres=rep)
    assert loader.num() == 5
    for i in range(5):
        s = next(loader)
        assert s.duration() == 0.5

def test_audio_select_loader_with_attrs(five_time_stamped_wave_files):
    """ Test that we can use the AudioSelectionLoader class to compute MagSpectrograms with 
        extra attributes from selection table""" 
    rep = {'type':'MagSpectrogram','window':0.1,'step':0.02}
    # create a selection table
    files = find_wave_files(path=five_time_stamped_wave_files, return_path=False, search_subdirs=True)
    sel = pd.DataFrame({'filename':[files[0],files[1]],
                        'start':[0.10,0.12],
                        'end':[0.46,0.42],
                        'comment':['big','small'],
                        'conf':[0.31,0.99]})
    sel = use_multi_indexing(sel, 'sel_id')
    # init loader all attrs
    loader = AudioSelectionLoader(path=five_time_stamped_wave_files, selections=sel, repres=rep, include_attrs=True)
    s = next(loader)
    attrs = s.get_instance_attrs()
    assert attrs['comment'] == 'big'
    assert attrs['conf'] == 0.31
    s = next(loader)
    attrs = s.get_instance_attrs()
    assert attrs['comment'] == 'small'
    assert attrs['conf'] == 0.99
    # init loader, selected attrs
    loader = AudioSelectionLoader(path=five_time_stamped_wave_files, selections=sel, repres=rep, attrs=['comment','dummy'])
    s = next(loader)
    attrs = s.get_instance_attrs()
    assert attrs['comment'] == 'big'
    assert 'conf' not in attrs.keys()
    assert 'dummy' not in attrs.keys()

def test_audio_select_loader_accepts_kwargs(five_time_stamped_wave_files):
    """ Test that we can use the AudioSelectionLoader class to compute MagSpectrograms with 
        a complex phase""" 
    rep = {'type':'MagSpectrogram','window':0.1,'step':0.02}
    # create a selection table
    files = find_wave_files(path=five_time_stamped_wave_files, return_path=False, search_subdirs=True)
    sel = pd.DataFrame({'filename':[files[0],files[1]],
                        'start':[0.10,0.12],
                        'end':[0.46,0.42],
                        'comment':['big','small'],
                        'conf':[0.31,0.99]})
    sel = use_multi_indexing(sel, 'sel_id')
    # init loader all attrs
    loader = AudioSelectionLoader(path=five_time_stamped_wave_files, selections=sel, repres=rep, compute_phase=True)
    s = next(loader)
    assert np.ndim(s.data) == 3

def test_audio_select_loader_start_end_outside_file(five_time_stamped_wave_files):
    """ Test that we can use the AudioSelectionLoader class to compute MagSpectrograms
        for selections that extend beyond the limits of the wav file """ 
    rep = {'type':'MagSpectrogram','window':0.02,'step':0.01}
    # create a selection table
    files = find_wave_files(path=five_time_stamped_wave_files, return_path=False, search_subdirs=True)
    sel = pd.DataFrame({'filename':[files[0],files[1],files[3]],
                        'start':[-0.20,0.12,0.44],
                        'end':[0.11,0.43,0.75]})
    sel = use_multi_indexing(sel, 'sel_id')
    # init loader
    loader = AudioSelectionLoader(path=five_time_stamped_wave_files, selections=sel, repres=rep)
    assert loader.num() == 3
    for i in range(3):
        s = next(loader)
        assert s.duration() == 0.31

def test_audio_frame_loader_get_files():
    """Test that the get_file_paths method of the AudioFrameLoader class works"""
    path = os.path.join(path_to_assets, 'wav_files')
    loader = AudioFrameLoader(path=path, frame=30., step=15.)
    file_paths = loader.get_file_paths()
    expected = [os.path.join(path, 'subf', 'w3.wav'), os.path.join(path, 'w1.wav'), os.path.join(path, 'w2.wav')]
    assert file_paths == expected

def test_audio_frame_loader_in_batches_with_transforms(growing_sine_wave_file):
    """ Test that transform are applied correctly to audio loaded in batches""" 
    norm_trans = {'name':'normalize','mean':0.5,'std':1.2}
    rep = {'type':'Waveform', 'transforms':[norm_trans]}
    loader = AudioFrameLoader(filename=growing_sine_wave_file, frame=0.04, repres=rep, batch_size=2)
    wf1_b = next(loader)
    wf2_b = next(loader)
    loader = AudioFrameLoader(filename=growing_sine_wave_file, frame=0.04, repres=rep, batch_size=1)
    wf1 = next(loader)
    wf2 = next(loader)
    assert np.all(wf1_b.get_data() == wf1.get_data())
    assert np.all(wf2_b.get_data() == wf2.get_data())



import pytest
import numpy as np
import tensorflow as tf
from ketos.neural_networks.dev_utils.nn_interface import RecipeCompat
from ketos.neural_networks.densenet import ConvBlock, DenseBlock, TransitionBlock, DenseNetArch, DenseNetInterface
#from ketos.neural_networks.dev_utils.losses import FScoreLoss
from ketos.data_handling.data_feeding import BatchGenerator
#from ketos.neural_networks.dev_utils.metrics import Precision, Recall, Accuracy, FScore
import os
import tables
import json

current_dir = os.path.dirname(os.path.realpath(__file__))
path_to_assets = os.path.join(os.path.dirname(current_dir),"assets")
path_to_tmp = os.path.join(path_to_assets,'tmp')


@pytest.fixture
def recipe_dict():
    recipe = {'interface': 'DenseNetInterface',
                    'dense_blocks':[ 6, 12, 24, 16],
                    'growth_rate':32,
                    'compression_factor':0.5,
                    'n_classes':2,
                    'dropout_rate':0.2,
                    'optimizer': {'recipe_name':'Adam', 'parameters': {'learning_rate':0.005}},
                    'loss_function': {'recipe_name':'CategoricalCrossentropy', 'parameters':{}},  
                    'metrics': [{'recipe_name':'CategoricalAccuracy', 'parameters':{}}]
        
    }
    return recipe

@pytest.fixture
def recipe():
    recipe = {'interface': 'DenseNetInterface',
                    'dense_blocks':[ 6, 12, 24, 16],
                    'growth_rate':32,
                    'compression_factor':0.5,
                    'n_classes':2,
                    'dropout_rate':0.2,
                    'optimizer': RecipeCompat('Adam', tf.keras.optimizers.Adam, learning_rate=0.005),
                    'loss_function': RecipeCompat('CategoricalCrossentropy', tf.keras.losses.CategoricalCrossentropy),  
                    'metrics': [RecipeCompat('CategoricalAccuracy',tf.keras.metrics.CategoricalAccuracy),
                                ],
                    }
    return recipe


def test_ConvBlock():
    block = ConvBlock(growth_rate=32)

    assert len(block.layers) == 6

    assert isinstance(block.layers[0], tf.keras.layers.BatchNormalization)
    assert isinstance(block.layers[1], tf.keras.layers.Activation)
    assert isinstance(block.layers[2], tf.keras.layers.Conv2D)

    assert isinstance(block.layers[3], tf.keras.layers.BatchNormalization)
    assert isinstance(block.layers[4], tf.keras.layers.Activation)
    assert isinstance(block.layers[5], tf.keras.layers.Conv2D)


def test_DenseBlock():
    block = DenseBlock(growth_rate=32, n_blocks=16)

    assert len(block.layers) == 1
    assert len(block.layers[0].layers) == 16
    for i in range(16):
        assert isinstance(block.layers[0].layers[i], ConvBlock)
        assert len(block.layers[0].layers[i].layers) == 6
        assert isinstance(block.layers[0].layers[i].layers[0], tf.keras.layers.BatchNormalization)
        assert isinstance(block.layers[0].layers[i].layers[1], tf.keras.layers.Activation)
        assert isinstance(block.layers[0].layers[i].layers[2], tf.keras.layers.Conv2D)

        assert isinstance(block.layers[0].layers[i].layers[3], tf.keras.layers.BatchNormalization)
        assert isinstance(block.layers[0].layers[i].layers[4], tf.keras.layers.Activation)
        assert isinstance(block.layers[0].layers[i].layers[5], tf.keras.layers.Conv2D)


def test_TransitionBlock():
    block = TransitionBlock(n_channels=64, compression_factor=0.5)

    assert len(block.layers) == 5
    assert isinstance(block.layers[0], tf.keras.layers.BatchNormalization)
    assert isinstance(block.layers[1], tf.keras.layers.Conv2D)
    assert isinstance(block.layers[2], tf.keras.layers.Dropout)
    assert isinstance(block.layers[3], tf.keras.layers.Activation)
    assert isinstance(block.layers[4], tf.keras.layers.AveragePooling2D)

def test_DenseNetArch(recipe):
    densenet = DenseNetArch(dense_blocks=recipe['dense_blocks'], growth_rate=recipe['growth_rate'],
                            compression_factor=recipe['compression_factor'], n_classes=recipe['n_classes'],
                            dropout_rate=recipe['dropout_rate'])

    
    assert len(densenet.layers) == 9

    assert isinstance(densenet.layers[0], tf.keras.layers.Conv2D)
    assert isinstance(densenet.layers[1], tf.keras.layers.BatchNormalization)
    assert isinstance(densenet.layers[2], tf.keras.layers.Activation)
    assert isinstance(densenet.layers[3], tf.keras.layers.MaxPool2D)

    #Dense blocks
    assert len(densenet.layers[4].layers) == (len(recipe['dense_blocks']) * 2)
    print(densenet.layers[4].layers)
    for block_id, block in enumerate(densenet.layers[4].layers):
        if block_id % 2 == 0: #DenseBlock
            assert isinstance(block, DenseBlock)
            assert len(block.layers[0].layers) == recipe['dense_blocks'][int(block_id / 2)]
        else: #TransitionBlock     
            assert isinstance(block, TransitionBlock)


    
def test_DenseNet_Interface_build_from_recipe(recipe):
    densenet = DenseNetInterface._build_from_recipe(recipe)

    assert densenet.optimizer.recipe_name == recipe['optimizer'].recipe_name
    assert densenet.optimizer.instance.__class__ == recipe['optimizer'].instance.__class__
    assert densenet.optimizer.args == recipe['optimizer'].args

    assert densenet.loss_function.recipe_name == recipe['loss_function'].recipe_name
    assert densenet.loss_function.instance.__class__ == recipe['loss_function'].instance.__class__
    assert densenet.loss_function.args == recipe['loss_function'].args

    assert densenet.metrics[0].recipe_name == recipe['metrics'][0].recipe_name
    assert densenet.metrics[0].instance.__class__ == recipe['metrics'][0].instance.__class__
    assert densenet.metrics[0].args == recipe['metrics'][0].args

    assert densenet.dense_blocks == recipe['dense_blocks']
    assert densenet.growth_rate == recipe['growth_rate']
    assert densenet.compression_factor ==  recipe['compression_factor']    
    assert densenet.dropout_rate ==  recipe['dropout_rate']    


def test_extract_recipe_dict(recipe, recipe_dict):
    densenet = DenseNetInterface._build_from_recipe(recipe)
    written_recipe = densenet._extract_recipe_dict()

    assert written_recipe == recipe_dict

def test_read_recipe_file(recipe, recipe_dict):
    path_to_recipe_file = os.path.join(path_to_tmp, "test_densenet_recipe.json")
    densenet = DenseNetInterface._build_from_recipe(recipe)
    written_recipe = densenet._extract_recipe_dict()
    densenet.save_recipe_file(path_to_recipe_file)

    #Read recipe as a recipe dict
    read_recipe = densenet._read_recipe_file(path_to_recipe_file,return_recipe_compat=False)
    assert read_recipe == recipe_dict

    #Read recipe as a recipe dict with RecipeCompat objects
    read_recipe = densenet._read_recipe_file(path_to_recipe_file,return_recipe_compat=True)
    assert read_recipe['optimizer'].recipe_name ==recipe['optimizer'].recipe_name
    assert read_recipe['optimizer'].instance.__class__ == recipe['optimizer'].instance.__class__
    assert read_recipe['optimizer'].args == recipe['optimizer'].args

    assert read_recipe['loss_function'].recipe_name == recipe['loss_function'].recipe_name
    assert read_recipe['loss_function'].instance.__class__ == recipe['loss_function'].instance.__class__
    assert read_recipe['loss_function'].args == recipe['loss_function'].args
    
    assert read_recipe['metrics'][0].recipe_name == recipe['metrics'][0].recipe_name
    assert read_recipe['metrics'][0].instance.__class__ == recipe['metrics'][0].instance.__class__
    assert read_recipe['metrics'][0].args == recipe['metrics'][0].args

    assert read_recipe['dense_blocks'] == recipe['dense_blocks']
    assert read_recipe['growth_rate'] == recipe['growth_rate']
    assert read_recipe['compression_factor'] ==  recipe['compression_factor']    
    assert read_recipe['dropout_rate'] ==  recipe['dropout_rate']  


def test_train_DenseNet(sample_data):
    data, labels = sample_data
    densenet = DenseNetInterface() #default densenet
    train_generator = BatchGenerator(batch_size=5, x=data, y=labels, shuffle=True)
    val_generator = BatchGenerator(batch_size=5, x=data, y=labels, shuffle=True)

    densenet.train_generator = train_generator
    densenet.val_generator = val_generator

    densenet.train_loop(2, checkpoint_freq=0)





# ================================================================================ #
#   Authors: Fabio Frazao and Oliver Kirsebom                                      #
#   Contact: fsfrazao@dal.ca, oliver.kirsebom@dal.ca                               #
#   Organization: MERIDIAN (https://meridian.cs.dal.ca/)                           #
#   Team: Data Analytics                                                           #
#   Project: ketos                                                                 #
#   Project goal: The ketos library provides functionalities for handling          #
#   and processing acoustic data and applying deep neural networks to sound        #
#   detection and classification tasks.                                            #
#                                                                                  #
#   License: GNU GPLv3                                                             #
#                                                                                  #
#       This program is free software: you can redistribute it and/or modify       #
#       it under the terms of the GNU General Public License as published by       #
#       the Free Software Foundation, either version 3 of the License, or          #
#       (at your option) any later version.                                        #
#                                                                                  #
#       This program is distributed in the hope that it will be useful,            #
#       but WITHOUT ANY WARRANTY; without even the implied warranty of             #
#       MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the              #
#       GNU General Public License for more details.                               # 
#                                                                                  #
#       You should have received a copy of the GNU General Public License          #
#       along with this program.  If not, see <https://www.gnu.org/licenses/>.     #
# ================================================================================ #

""" 'neural_networks.dev_utils.export' module within the ketos library

    This module contains utilities for saving ketos models in various formats.
"""

from ketos.data_handling.parsing import parse_audio_representation, encode_audio_representation, parse_parameter, encode_parameter
from tensorflow.saved_model import save as save_pb
from ketos.audio.audio_loader import audio_repres_dict
from zipfile import ZipFile
import warnings
import json
import shutil
import os


def export_to_ketos_protobuf(model, output_name, input_shape, audio_repr=None, audio_repr_file=None, 
                            tmp_folder="tmp_export_folder", overwrite=False, input_duration=None):
    """ Export a ketos model to Ketos-Protobuf format. Protobuf is a free and open-source 
        cross-platform data format developed by Google. 

        Saving your ketos model in Ketos-Protobuf format makes it easier to share it with 
        collaborators and use it with other software applications.

        In particular, the output file generated by this function can be loaded directly into 
        PAMGuard, an open-source and widely adopted application for passive acoustic monitoring (PAM).

        The function generates a zipped archive containing,
        
         * the tensorflow model in protobuf format (model.pb)
         * the audiot representation (audio_repr.json)
         * the ketos model recipe (recipe.json)
       
        The user is free to specify the extension of the output file, but we recommend using \*.ktpb
        as this will allow the file to be recognized and loaded into PAMGuard.

        Args:
            model: 
                The ketos model to be exported. Usually created by one of the Interface classes found 
                in ketos.neural_networks (e.g.: ResNetInterface)
            output_name: str
                The name of the exported model. Must have the extension '.ktpb" to ensure that it can 
                be loaded into PAMGuard.
            input_shape: list or tuple.
                The input shape expected by the model. It can be represented by a tuple or list of four elements: 
                [number of intances, width, height, number of channels). The number of instances and number of channels 
                are commonly 1, and the width and height are usually the number of time and frequency bins ins a 
                spectrogram, respectively. This, however, can vary with the model in question.
            audio_repr: dict
                Audio representation. For example,
                    
                >>> audio_repr = {"spectrogram": {
                ...                   "type": "MagSpectrogram",
                ...                   "rate": "1000 Hz", 
                ...                   "window": "0.256 s",
                ...                   "step": "0.032 s",
                ...                   "freq_min": "0 Hz",
                ...                   "freq_max": "500 Hz",
                ...                   "window_func": "hamming",
                ...                   "transforms": [{"name":"normalize"}]}
                ...              }                

            audio_repr_file: str
                Path to json file containing audio representation. Overwrites audio_repr, if specified.
            tmp_folder: str
                The name for a temporary folder created during the model conversion. It will be deleted upon sucessful execution. 
                If the folder already exists, a 'FileExistsError will be thrown ( unless 'overwite' is set to True).
            overwrite: bool    
                If true and the folder specified in 'tmp_folder' exists, the folder will be overwritten.
            input_duration: None or float
                Duration in seconds of the input sample. If input_duration is None, the duration is taken from the 
                audio representation, or, if not available there, computed as step * input_shape[1]
    """
    assert audio_repr is not None or audio_repr_file is not None, 'either audio_repr or audio_repr_file must be specified'

    if audio_repr_file is not None:
        with open(audio_repr_file, 'r') as fil:
            audio_repr = json.load(fil)
        
    if os.path.exists(tmp_folder):
        if not overwrite:
            raise FileExistsError("{} already exists. If you want to overwrite it set the 'overwrite' argument to True.".format(tmp_folder))
        else:
            shutil.rmtree(tmp_folder)

    assert model.model.built, "The model must be built. Call model.run_on_instance() on a sample input"

    # loop over audio representations (usually, there will only be 1)
    for ar in audio_repr.values():

        assert 'step' in ar.keys(), 'step parameter is missing from audio representation'

        # encode audio representation if it isn't already encoded
        if isinstance(ar['step'], (int, float)): 
            ar = encode_audio_representation(ar)

        if input_duration is None:
            if 'duration' not in ar:
                duration = parse_parameter(name='step', value=ar['step']) * input_shape[1]
                ar['duration'] = encode_parameter(name='duration', value=duration)

        else:
            ar['duration'] = encode_parameter(name='duration', value=input_duration)

        ar['dtype'] = model.model.dtype
        ar['input_ndims'] = model.model.layers[0].input_spec.min_ndim
        ar['input_shape'] = input_shape

    os.makedirs(tmp_folder)
    recipe_path = os.path.join(tmp_folder, 'recipe.json')
    model.save_recipe_file(recipe_path)
    model_path = os.path.join(tmp_folder, 'model')
    save_pb(obj=model.model, export_dir=model_path)

    with ZipFile(output_name, 'w') as zip:
        zip.write(model_path, "model")

        for root, dirs, files in os.walk(model_path):
            renamed_root = root.replace(model_path, "model")
            for d in dirs:
                zip.write(os.path.join(root,d), os.path.join(renamed_root,d))
            for f in files:
                zip.write(os.path.join(root,f),os.path.join(renamed_root,f))            
                    
        zip.write(recipe_path, "recipe.json")

        audio_repr_path = os.path.join(tmp_folder, "audio_repr.json")
        with open(audio_repr_path, 'w') as json_repr:
            json.dump(audio_repr, json_repr)

        zip.write(audio_repr_path, "audio_repr.json")
        
    shutil.rmtree(tmp_folder)


def export_to_protobuf(model, output_folder):
    """ Export a ketos model to Protobuf format (\*.pb).

        See also the related fuction :func:`export_to_ketos_protobuf`.

        Args:
            model: 
                The ketos model to be exported. Usually created by one of the Interface classes found 
                in ketos.neural_networks (e.g.: ResNetInterface)
            output_folder: str
                Folder where the exported model will be saved
    """
    assert model.model.built, "The model must be built. Call model.run_on_instance() on a sample input"

    save_pb(obj=model.model, export_dir=output_folder)


from .cnn import CNNInterface, CNN1DInterface
from .resnet import ResNetInterface, ResNet1DInterface
from .densenet import DenseNetInterface
from .inception import InceptionInterface
from zipfile import ZipFile
from shutil import rmtree
import json
import os


interface_names_in_recipes = {'CNNInterface':CNNInterface,
                              'CNN1DInterface':CNN1DInterface,
                              'ResNetInterface':ResNetInterface,
                              'ResNet1DInterface':ResNet1DInterface,
                              'DenseNetInterface':DenseNetInterface,
                              'InceptionInterface':InceptionInterface}

def load_model_file(model_file, new_model_folder,  overwrite=True, load_audio_repr=False,  replace_top=False, diff_n_classes=None):
        """ Load a model from a ketos (.kt) model file.

            Args:
                model_file:str
                    Path to the ketos(.kt) file
                new_model_folder:str
                    Path to folder where files associated with the model will be stored.
                overwrite: bool
                    If True, the 'new_model_folder' will be overwritten.
                replace_top: bool
                    If True, the classification top of the model will be replaced by a new, untrained one.
                    What is actually replaced (i.e.: what exactly is the "top") is defined by the architecture.
                    It is usually a block of dense layers with the appropriate activations. Default is False.
                diff_n_classes: int
                    Only relevant when 'replace_top' is True.
                    If the new model should have a different number of classes it can be specified by this parameter.
                    If left to none, the new model will have the same number of classes as the original.
                load_audio_repr: bool
                    If True, look for an audio representation included with the model. 
                    
            Raises:
                FileExistsError: If the 'new_model_folder' already exists and 'overwite' is False.
                ValueError: If the model recipe does not contain a valid value for the 'interface' field.

            Returns:
                model_instance: The loaded model
                audio_repr: If load_audio_repr is True, also return a dictionary with the loaded audio representation.

        """

        try:
            os.makedirs(new_model_folder)
        except FileExistsError:
            if overwrite == True:
                rmtree(new_model_folder)
                os.makedirs(new_model_folder)
            else:
                raise FileExistsError("Ketos needs a new folder for this model. Choose a folder name that does not exist or set 'overwrite' to True to replace the existing folder")

        with ZipFile(model_file, 'r') as zip:
            zip.extractall(path=new_model_folder)
    
        with open(os.path.join(new_model_folder,"recipe.json"), 'r') as json_recipe:
            recipe_dict = json.load(json_recipe)

        try:
            nn_interface = recipe_dict['interface']
            loaded = interface_names_in_recipes[nn_interface].load_model_file(model_file=model_file, new_model_folder=new_model_folder,  overwrite=True, load_audio_repr=load_audio_repr,  replace_top=replace_top, diff_n_classes=diff_n_classes)
        except KeyError:
            raise ValueError("The model recipe does not indicate a valid interface")

        return loaded
            


# ================================================================================ #
#   Authors: Fabio Frazao and Oliver Kirsebom                                      #
#   Contact: fsfrazao@dal.ca, oliver.kirsebom@dal.ca                               #
#   Organization: MERIDIAN (https://meridian.cs.dal.ca/)                           #
#   Team: Data Analytics                                                           #
#   Project: ketos                                                                 #
#   Project goal: The ketos library provides functionalities for handling          #
#   and processing acoustic data and applying deep neural networks to sound        #
#   detection and classification tasks.                                            #
#                                                                                  #
#   License: GNU GPLv3                                                             #
#                                                                                  #
#       This program is free software: you can redistribute it and/or modify       #
#       it under the terms of the GNU General Public License as published by       #
#       the Free Software Foundation, either version 3 of the License, or          #
#       (at your option) any later version.                                        #
#                                                                                  #
#       This program is distributed in the hope that it will be useful,            #
#       but WITHOUT ANY WARRANTY; without even the implied warranty of             #
#       MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the              #
#       GNU General Public License for more details.                               # 
#                                                                                  #
#       You should have received a copy of the GNU General Public License          #
#       along with this program.  If not, see <https://www.gnu.org/licenses/>.     #
# ================================================================================ #

""" 'audio.gammatone' module within the ketos library.

    This module provides utilities to process audio data 
    using a auditory filterbank based on the gammatone function.

    Contents:
        GammatoneFilterBank class:
        AuralFeatures class
"""
import sys
import os
import copy
import librosa
import warnings
import numpy as np
from version_parser.version import Version
import scipy
import matplotlib.pyplot as plt
from ketos.audio.waveform import Waveform
from ketos.audio.annotation import AnnotationHandler
from ketos.audio.base_audio import BaseAudio, BaseAudioTime, segment_data


gammatone_filter_coeff = dict()


def compute_center_freqs(num_chan, sampl_rate, freq_min):
    """ Compute the center frequencies of the filter bank

        Args:
            num_chan: int
                Number of channels in the filter bank
            sampl_rate: float
                Sampling rate in Hz
            freq_min: float
                Minimum frequency in Hz

        Returns:
            freqs: float
                Center frequencies in Hz
    """
    f0 = 228.8
    i = np.linspace(1, num_chan, num_chan)
    freqs = np.exp(i / num_chan * (np.log(freq_min + f0) - np.log(sampl_rate / 2. + f0))) * (sampl_rate / 2. + f0) - f0
    freqs = np.flip(freqs)
    return freqs

def filter_signal(signal, sampl_rate, freqs):
    """ Pass the signal through the gammatone filters

        Args:
            signal: numpy.array
                Audio signal
            sampl_rate: float
                Sampling rate in Hz
            freqs: numpy.array
                Center frequencies of the filter bank

        Returns:
            x: numpy.array
                The filtered signals stacked vertically into 2D array
    """
    x = []
    for freq in freqs:
        (b, a) = get_filter_coeffs(sampl_rate, freq)
        x.append(scipy.signal.filtfilt(b, a, signal))

    x = np.array(x)
    x = np.swapaxes(x, 0, 1)
    return x

def apply_weight_func(x, freqs):
    """ Apply C weighting function.
    
        This weighting function represents the approximate frequency sensitivity 
        of the human auditory system.

        Args:
            x: numpy.array
                The filtered signals
            freqs: numpy.array
                Center frequencies of the filter bank

        Returns:
            x: numpy.array
                The C-weighted filtered signals
    """
    C = 1.007 * 12200**2 * freqs**2 / (freqs**2 + 20.6**2) / (freqs**2 + 12200**2)
    return x * C[np.newaxis,:]

def get_filter_coeffs(sampl_rate, freq):
    """ Get the gammatone filter coefficients.
    
        Args:
            sampl_rate: float
                Sampling rate in Hz
            freq: float
                Center frequency in Hz

        Returns:
            : tuple
                Coefficients of the gammatone filter
    """        
    key = (sampl_rate, freq) 
    if key not in gammatone_filter_coeff.keys():
        gammatone_filter_coeff[key] = compute_filter_coeffs(sampl_rate, freq)

    return gammatone_filter_coeff[key]        
        
def compute_filter_coeffs(sampl_rate, freq):
    """ Compute the gammatone filter coefficients.
    
        Args:
            sampl_rate: float
                Sampling rate in Hz
            freq: float
                Center frequency in Hz

        Returns:
            (b, a): tuple
                Coefficients of the gammatone filter
    """        
    if Version(scipy.__version__) < Version("1.6.0"):
        print('The `compute_filter_coeffs` method in the `gammatone` module requires Scipy>=1.6.0')
        print(f'The present environment only has Scipy=={Version(scipy.__version__)}')
        print('Note that Scipy>=1.6.0 requires Python>=3.7.0')
        exit(1)

    b, a = scipy.signal.gammatone(freq=freq, ftype='iir', fs=sampl_rate)
    return (b, a)


class GammatoneFilterBank(BaseAudioTime):
    """ Gammatone filter bank.

        The filtered signals are stored in a 2D numpy array, where the first axis 
        (0) is the time dimension and the second axis (1) is the frequency dimension.

        Args:
            data: 2d numpy array
                Filtered data 
            rate: float
                Sampling rate in Hz
            freqs: array-like
                Center frequencies of the filter bank in Hz
            filename: str or list(str)
                Name of the source audio file, if available.   
            offset: float or array-like
                Position in seconds of the left edge of the spectrogram within the source 
                audio file, if available.
            label: int
                Spectrogram label. Optional
            annot: AnnotationHandler
                AnnotationHandler object. Optional
            weight_func: bool
                Apply C weighting function. Default is True.
            
        Attributes:
            data: 2d numpy array
                Filtered data 
            rate: float
                Sampling rate in Hz
            freqs: array-like
                Center frequencies of the filter bank in Hz
            filename: str or list(str)
                Name of the source audio file, if available.   
            offset: float or array-like
                Position in seconds of the left edge of the spectrogram within the source 
                audio file, if available.
            label: int
                Spectrogram label. Optional
            annot: AnnotationHandler
                AnnotationHandler object. Optional
            weight_func: bool
                Apply C weighting function.
"""
    def __init__(self, data, rate, freqs, filename=None, offset=0, label=None, annot=None, weight_func=True, **kwargs):
        self.rate = rate
        self.freqs = freqs
        self.weight_func = weight_func
        super().__init__(data=data, time_res=1./rate, filename=filename, offset=offset, label=label, annot=annot)

    @classmethod
    def empty(cls):
        """ Creates an empty GammatoneFilterBank object
        """
        return cls(data=np.empty(shape=(0,0), dtype=np.float64), rate=1, freqs=[])

    @classmethod
    def from_waveform(cls, audio, num_chan=20, freq_min=1, weight_func=True):
        """ Create a Gammatone Filter Bank from an instance of :class:`audio_signal.Waveform`.
        
            Args:
                audio: Waveform
                    Audio signal 
                num_chan: int
                    Number of channels in the filter bank
                freq_min: float
                    Minimum frequency of the filter bank in Hz
                weight_func: bool
                    Apply C weighting function. Default is True.

            Returns:
                gfb: GammatoneFilterBank
                    Gammatone filter bank
        """
        center_freqs = compute_center_freqs(num_chan=num_chan, sampl_rate=audio.rate, freq_min=freq_min)

        filtered_signals = filter_signal(signal=audio.data, sampl_rate=audio.rate, freqs=center_freqs)

        if weight_func:
            filtered_signals = apply_weight_func(x=filtered_signals, freqs=center_freqs)

        gfb = cls(data=filtered_signals, rate=audio.rate, freqs=center_freqs,
            filename=audio.filename, offset=audio.offset, label=audio.label, 
            annot=audio.annot, weight_func=weight_func)

        return gfb

    @classmethod
    def from_wav(cls, path, num_chan=20, freq_min=1, channel=0, rate=None, offset=0, duration=None,
            resample_method='scipy', id=None, normalize_wav=False, weight_func=True, **kwargs):
        """ Create a Gammatone Filter Bank directly from wav file.

            The arguments offset and duration can be used to select a portion of the wav file.
            
            Note that values specified for the arguments offset and duration may be subject 
            to slight adjustments to ensure that the selected portion corresponds to an integer 
            number of samples.

            Args:
                path: str
                    Path to wav file
                num_chan: int
                    Number of channels in the filter bank
                freq_min: float
                    Minimum frequency of the filter bank in Hz
                channel: int
                    Channel to read from. Only relevant for stereo recordings
                rate: float
                    Desired sampling rate in Hz. If None, the original sampling rate will be used
                offset: float
                    Start time of selection in seconds, relative the start of the wav file.
                duration: float
                    Length of selection in seconds.
                resample_method: str
                    Resampling method. Only relevant if `rate` is specified. Options are
                        * kaiser_best
                        * kaiser_fast
                        * scipy (default)
                        * polyphase

                    See https://librosa.github.io/librosa/generated/librosa.core.resample.html 
                    for details on the individual methods.
                id: str
                    Unique identifier (optional). If None, the filename will be used.
                normalize_wav: bool
                    Normalize the waveform to have a mean of zero (mean=0) and a standard 
                    deviation of unity (std=1) before computing the spectrogram. Default is False.
                weight_func: bool
                    Apply C weighting function. Default is True.

            Returns:
                : GammatoneFilterBank
                    Gammatone filter bank

            Example:
                >>> # load gammatone filter bank from wav file
                >>> from ketos.audio.gammatone import GammatoneFilterBank
                >>> gfb = GammatoneFilterBank.from_wav('ketos/tests/assets/grunt1.wav', num_chan=20, freq_min=10, rate=1000)
                >>> # print the center frequencies rounded to 1 decimal
                >>> print(np.round(gfb.freqs,1))
                [ 10.   23.7  38.2  53.5  69.7  86.8 104.9 124.1 144.3 165.7 188.4 212.3
                 237.6 264.4 292.7 322.6 354.2 387.7 423.1 460.5]
                >>> # display the 4th filter bank signal
                >>> fig = gfb.plot(filter_id=3)
                >>> fig.savefig("ketos/tests/assets/tmp/gfb3_grunt1.png")
                >>> plt.close(fig)

                .. image:: ../../../ketos/tests/assets/tmp/gfb3_grunt1.png
        """
        # load audio
        audio = Waveform.from_wav(path=path, channel=channel, rate=rate, offset=offset, duration=duration, 
            resample_method=resample_method, id=id, normalize_wav=normalize_wav, **kwargs)

        if len(audio.get_data()) == 0:
            warnings.warn("Empty GammatoneFilterBank returned", RuntimeWarning)
            return cls.empty()

        # compute gammatone filter bank
        return cls.from_waveform(audio=audio, num_chan=num_chan, freq_min=freq_min, weight_func=weight_func)

    def get_repres_attrs(self):
        """ Get audio representation attributes """ 
        attrs = super().get_repres_attrs()
        attrs.update({'rate':self.rate, 'freqs':self.freqs, 
            'weight_func':self.weight_func, 'type':self.__class__.__name__})
        return attrs

    def plot(self, filter_id, show_annot=False, figsize=(5,4), label_in_title=True, show_envelope=False):
        """ Plot the filtered signal with proper axes ranges and labels.

            Optionally, also display annotations as boxes superimposed on the signal.

            Note: The resulting figure can be shown (fig.show())
            or saved (fig.savefig(file_name))

            Args:
                filter_id: int
                    Filter to be plotted.
                show_annot: bool
                    Display annotations
                figsize: tuple
                    Figure size
                label_in_title: bool
                    Include label (if available) in figure title
                show_envelope: bool
                    Display envelope on top of signal
            
            Returns:
                : matplotlib.figure.Figure
                    A figure object.

            Example:
                >>> from ketos.audio.gammatone import GammatoneFilterBank
                >>> # load gammatone filter bank
                >>> gfb = GammatoneFilterBank.from_wav('ketos/tests/assets/grunt1.wav', num_chan=20, freq_min=10, rate=1000)
                >>> # add an annotation
                >>> gfb.annotate(start=1.2, end=1.6, freq_min=70, freq_max=600, label=1)
                >>> # show the 4th filter bank with annotation box
                >>> fig = gfb.plot(filter_id=3, show_annot=True)
                >>> fig.savefig("ketos/tests/assets/tmp/gfb3_w_annot_box.png")
                >>> plt.close(fig)

                .. image:: ../../../ketos/tests/assets/tmp/gfb3_w_annot_box.png
        """
        x = self.get_data()[:,filter_id] # select the filtered signal 

        wf = Waveform(data=x, rate=self.rate, filename=self.get_filename(), 
            offset=self.get_offset(), label=self.get_label(), annot=self.get_annotations())

        return wf.plot(show_annot=show_annot, figsize=figsize, label_in_title=label_in_title,
            append_title=f', {self.freqs[filter_id]:.1f} Hz', show_envelope=show_envelope)



class AuralFeatures(BaseAudio):
    """ Aural features computed with the aural-features package (https://pypi.org/project/aural-features/).

        Args:
            data: 1d numpy array
                Feature values
            filename: str
                Name of the source audio file, if available.   
            offset: float
                Position in seconds of the left edge of the audio segment within the source 
                audio file, if available.
            label: int
                Label. Optional
            annot: AnnotationHandler
                AnnotationHandler object. Optional
"""
    def __init__(self, data, filename=None, offset=0, label=None, annot=None, waveform_transform_log=None, **kwargs):
        super().__init__(data=data, filename=filename, offset=offset, label=label, annot=annot)

        if waveform_transform_log is None: waveform_transform_log = []
        self.waveform_transform_log = waveform_transform_log

    @classmethod
    def from_waveform(cls, audio, filter_pad_samples=64, global_km_window_seconds=0.25, 
        local_km_window_seconds=0.008, filter_n=100, filter_min_hz=50):
        """ Compute aural features from an instance of :class:`audio_signal.Waveform`.
        
            Args:
                audio: Waveform
                    Audio signal 
                filter_pad_samples: int
                    Number of samples used for padding
                global_km_window_seconds: float
                    Length of global KM window in seconds
                local_km_window_seconds: float
                    Length of local KM window in seconds
                filter_n: int
                    Number of filters
                filter_min_hz: float
                    Min filter frequency in Hz 

            Returns:
                : AuralFeatures
                    Aural features
        """
        if 'aural' not in sys.modules:
            try:
                import aural.meridian as au
            except ImportError:
                print('aural-features package not found.')
                print('aural-features is required by the AuralFeatures class.')
                print('install with `pip install aural-features`.')
                print('note that aural-features requires Scipy>=1.6 and Python>=3.7')
                raise ImportError

        conf = au.Config() # For defaults, should be safe to start

        conf.filter_pad_samples = filter_pad_samples
        conf.global_km_window_seconds = global_km_window_seconds
        conf.local_km_window_seconds = local_km_window_seconds
        conf.filter_n = filter_n
        conf.filter_min_hz = filter_min_hz
        
        try:
            duration, features = au.extract(audio.get_data(), audio.rate, conf)

            values = [duration]
            for x in features:
                for name, value in x._asdict().items():
                    values.append(value)

            values = np.array(values)
            values = np.nan_to_num(values, nan=0.0) #replace NaN's with zeros

        except au.IsolationFailed:
            values = np.zeros(46)

        return cls(data=values, filename=audio.filename, offset=audio.offset, label=audio.label, annot=audio.annot)

    @classmethod
    def from_wav(cls, path, filter_pad_samples=64, global_km_window_seconds=0.25, 
            local_km_window_seconds=0.008, filter_n=100, filter_min_hz=50, 
            channel=0, rate=None, offset=0, duration=None,
            resample_method='scipy', id=None, normalize_wav=False, 
            waveform_transforms=None, **kwargs):
        """ Compute aural features directly from wav file.

            The arguments offset and duration can be used to select a portion of the wav file.
            
            Note that values specified for the arguments offset and duration may be subject 
            to slight adjustments to ensure that the selected portion corresponds to an integer 
            number of samples.

            Args:
                path: str
                    Path to wav file
                filter_min: float
                    Min filter frequency in Hz 
                local_km_window: float
                    Length of local KM window in seconds
                channel: int
                    Channel to read from. Only relevant for stereo recordings
                rate: float
                    Desired sampling rate in Hz. If None, the original sampling rate will be used
                offset: float
                    Start time of selection in seconds, relative the start of the wav file.
                duration: float
                    Length of selection in seconds.
                resample_method: str
                    Resampling method. Only relevant if `rate` is specified. Options are
                        * kaiser_best
                        * kaiser_fast
                        * scipy (default)
                        * polyphase

                    See https://librosa.github.io/librosa/generated/librosa.core.resample.html 
                    for details on the individual methods.
                id: str
                    Unique identifier (optional). If None, the filename will be used.
                normalize_wav: bool
                    Normalize the waveform to have a mean of zero (mean=0) and a standard 
                    deviation of unity (std=1) before computing the spectrogram. Default is False.
                waveform_transforms: list(dict)
                    List of dictionaries, where each dictionary specifies the name of 
                    a transformation to be applied to the waveform before generating 
                    the spectrogram. For example,
                    {"name":"add_gaussian_noise", "sigma":0.5}

            Returns:
                : AuralFeatures
                    Aural features
        """
        # load audio
        audio = Waveform.from_wav(path=path, channel=channel, rate=rate, offset=offset, duration=duration, 
            resample_method=resample_method, id=id, normalize_wav=normalize_wav, 
            transforms=waveform_transforms, **kwargs)

        if len(audio.get_data()) == 0:
            warnings.warn("Empty AuralFeatures returned", RuntimeWarning)
            return cls.empty()

        # compute gammatone filter bank
        return cls.from_waveform(audio=audio, filter_pad_samples=filter_pad_samples, global_km_window_seconds=global_km_window_seconds, 
            local_km_window_seconds=local_km_window_seconds, filter_n=filter_n, filter_min_hz=filter_min_hz)

    def get_repres_attrs(self):
        """ Get audio representation attributes """ 
        attrs = super().get_repres_attrs()
        attrs.update({'type':self.__class__.__name__})
        return attrs



# ================================================================================ #
#   Authors: Fabio Frazao and Oliver Kirsebom                                      #
#   Contact: fsfrazao@dal.ca, oliver.kirsebom@dal.ca                               #
#   Organization: MERIDIAN (https://meridian.cs.dal.ca/)                           #
#   Team: Data Analytics                                                           #
#   Project: ketos                                                                 #
#   Project goal: The ketos library provides functionalities for handling          #
#   and processing acoustic data and applying deep neural networks to sound        #
#   detection and classification tasks.                                            #
#                                                                                  #
#   License: GNU GPLv3                                                             #
#                                                                                  #
#       This program is free software: you can redistribute it and/or modify       #
#       it under the terms of the GNU General Public License as published by       #
#       the Free Software Foundation, either version 3 of the License, or          #
#       (at your option) any later version.                                        #
#                                                                                  #
#       This program is distributed in the hope that it will be useful,            #
#       but WITHOUT ANY WARRANTY; without even the implied warranty of             #
#       MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the              #
#       GNU General Public License for more details.                               # 
#                                                                                  #
#       You should have received a copy of the GNU General Public License          #
#       along with this program.  If not, see <https://www.gnu.org/licenses/>.     #
# ================================================================================ #

""" Unit tests for the 'neural_networks.dev_utils.export' module within the ketos library
"""
import pytest
import os
import numpy as np
from ketos.neural_networks import load_model_file
from ketos.audio.audio_loader import audio_repres_dict
from ketos.neural_networks.dev_utils.export import export_to_protobuf, export_to_ketos_protobuf


current_dir = os.path.dirname(os.path.realpath(__file__))
path_to_assets = os.path.join(os.path.dirname(current_dir),"assets")
path_to_tmp = os.path.join(path_to_assets,'tmp')


def test_export_to_ketos_protobuf():
    """Test export resnet to pamguard format"""
    model_path = os.path.join(path_to_assets, 'narw_resnet.kt')
    tmp_path = os.path.join(path_to_tmp, 'tmp_folder')
    model, audio_repr = load_model_file(model_path, tmp_path, load_audio_repr=True)
    input_spec = np.ones(shape=(94,129))
    model.run_on_instance(input_spec)
    output_path = os.path.join(path_to_tmp, 'narw1.ktpb')
    export_to_ketos_protobuf(model=model, output_name=output_path, audio_repr=audio_repr[0], overwrite=True, input_shape=(1,94,129,1))
    assert os.path.isfile(output_path)

def test_export_to_ketos_protobuf_audio_repr_file():
    """Test export resnet to pamguard format using audio representation file path"""
    model_path = os.path.join(path_to_assets, 'narw_resnet.kt')
    audio_repr_path = os.path.join(path_to_assets, 'audio_repr.json')
    tmp_path = os.path.join(path_to_tmp, 'tmp_folder')
    model, audio_repr = load_model_file(model_path, tmp_path, load_audio_repr=True)
    input_spec = np.ones(shape=(94,129))
    model.run_on_instance(input_spec)
    output_path = os.path.join(path_to_tmp, 'narw2.ktpb')
    export_to_ketos_protobuf(model=model, output_name=output_path, audio_repr_file=audio_repr_path, overwrite=True, input_shape=(1,94,129,1))
    assert os.path.isfile(output_path)

def test_export_to_protobuf():
    """Test export resnet to protobuf format"""
    model_path = os.path.join(path_to_assets, 'narw_resnet.kt')
    tmp_path = os.path.join(path_to_tmp, 'tmp_folder')
    model = load_model_file(model_path, tmp_path, load_audio_repr=False)
    input_spec = np.ones(shape=(94,129))
    model.run_on_instance(input_spec)
    output_path = os.path.join(path_to_tmp, 'pb_model')
    export_to_protobuf(model=model, output_folder=output_path)
    assert os.path.isdir(output_path)



# ================================================================================ #
#   Authors: Fabio Frazao and Oliver Kirsebom                                      #
#   Contact: fsfrazao@dal.ca, oliver.kirsebom@dal.ca                               #
#   Organization: MERIDIAN (https://meridian.cs.dal.ca/)                           #
#   Team: Data Analytics                                                           #
#   Project: ketos                                                                 #
#   Project goal: The ketos library provides functionalities for handling          #
#   and processing acoustic data and applying deep neural networks to sound        #
#   detection and classification tasks.                                            #
#                                                                                  #
#   License: GNU GPLv3                                                             #
#                                                                                  #
#       This program is free software: you can redistribute it and/or modify       #
#       it under the terms of the GNU General Public License as published by       #
#       the Free Software Foundation, either version 3 of the License, or          #
#       (at your option) any later version.                                        #
#                                                                                  #
#       This program is distributed in the hope that it will be useful,            #
#       but WITHOUT ANY WARRANTY; without even the implied warranty of             #
#       MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the              #
#       GNU General Public License for more details.                               # 
#                                                                                  #
#       You should have received a copy of the GNU General Public License          #
#       along with this program.  If not, see <https://www.gnu.org/licenses/>.     #
# ================================================================================ #

""" detection sub-module within the ketos.neural_networks.dev_utils module

    This module provides auxiliary functions to incorporate pre-trained ketos classifiers models into detection tools

    Contents:
        

"""

import numpy as np
import pandas as pd
import os
from tqdm import tqdm
from ketos.audio.audio_loader import AudioLoader
from ketos.data_handling.data_feeding import BatchGenerator

def compute_avg_score(score_vector, win_len):
    """ Compute a moving average of the score vector.

        Args:
            score_vector: numpy array
                1d numpy array containing the target class classification score for each input
            win_len:int
                The window length for the moving average. Must be an odd integer

        Returns:
            numpy arrays
                One numpy arrays with the average scores each time step
    """
    assert isinstance(win_len, int) and win_len%2 == 1, 'win_len must be an odd integer'

    num_pad = int((win_len - 1) / 2)
    x = score_vector.astype(float)

    num_frames = len(score_vector) - 2*num_pad
    indices = np.tile(np.arange(0, win_len), (num_frames, 1)) + np.tile(np.arange(0, num_frames, 1), (win_len, 1)).T
    frames = x[indices.astype(np.int32, copy=False)]

    avg_score = np.nanmean(frames, axis=1)

    avg_score = np.pad(avg_score, (num_pad, num_pad), 'constant', constant_values=(0, 0)) 

    return avg_score


def map_detection_to_time(det_start, det_end, batch_start_timestamp, batch_end_timestamp, step, spec_dur, buffer):
    """ Converts the start and end of a detection from the position in the scores vecotr to time.

        Args:
            det_start: int
                The detection start expressed as an index in the scores vector.
            det_end: int
                The detection end (the last score that is part of the detection) expressed as an index in the scores vector. 
            batch_start_timestap:float
                The timestamp (in seconds from the beginning of the file) of the first score in the scores vector
                (i.e.: the score of the first input spectrogram in that batch)
            batch_end_timestap:float
                The timestamp (in seconds from the beginning of the file) of the last score in the scores vector
                (i.e.: the score of the last input spectrogram in that batch)
            step: float
                The time interval(in seconds) between the starts of each contiguous input spectrogram.
                For example, a step=0.5 indicates that the first spectrogram starts at time 0.0s (from the beginning of the audio file), the second at 0.5s, etc.
            spec_dur: float
                The duration of each spectrogram in seconds
            buffer: float
                Time (in seconds) to be added around the detection.

        Raises:
            ValueError: 
                If det_end is lower than det_start

        Returns:
            time_start, duration:float
                The corresponding start (in seconds from the beggining of the file) and duration
    """
    if det_end < det_start:
        raise ValueError("'det_end' cannot be lower than 'det_start'")

    time_start =  det_start * step - buffer 
    time_start += batch_start_timestamp
    time_start = max(batch_start_timestamp, time_start)
    
    time_end =  det_end * step + buffer 
    time_end += batch_start_timestamp
    time_end = min(batch_end_timestamp, time_end)

    duration = (time_end + step) - time_start
    
    return time_start, duration
    

def group_detections(scores_vector, batch_support_data, buffer=0.0, step=0.5, spec_dur=3.0, threshold=0.5):
    """ Groups time steps with a detection score above the specified threshold.

        Consecutive detections are grouped into one single detection represented by the 
        time interval (start-end, in seconds from beginning of the file).

        Args:
            scores_vector: numpy array
                1d numpy array containing the target class classification score for each input
            batch_support_data: numpy array
                An array of shape n x 2, where n is the batch size. The second dimension contains the filename and the start timestamp for each input in the batch
            buffer: float
                Time (in seconds) to be added around the detection
            step: float
                The time interval(in seconds) between the starts of each contiguous input spectrogram.
                For example, a step=0.5 indicates that the first spectrogram starts at time 0.0s (from the beginning of the audio file), the second at 0.5s, etc.
            spec_dur: float
                The duration of each spectrogram in seconds
            threshold: float or list of floats
                Minimum score value for a time step to be considered as a detection.
                
        Returns:
            det_timestamps: list of tuples
                The detections time stamp. Each item in the list is a tuple with the filename, start time, duration and score for that detection.
                The filename corresponds to the file where the detection started. 
                If a list of threshold values is specified, the returned object will be a list of lists with len(det_timestamps) = len(thresholds)         
    """
    thresholds = threshold if isinstance(threshold, list) else [threshold]

    det_timestamps = [[] for _ in range(len(thresholds))]

    for i in range(len(thresholds)):
        thres = thresholds[i]

        det_vector = np.where(scores_vector >= thres, 1.0, 0.0)
        within_det = False
        filename_vector = batch_support_data[:,0]

        prev_det_index = 0
        for det_index,det_value in enumerate(det_vector):

            is_new_file = (filename_vector[det_index] != filename_vector[max(0,prev_det_index)])
            is_end_of_vector = (det_index == len(det_vector) - 1)

            if det_value == 1.0 and not within_det: #start new detection
                start = det_index
                within_det = True
            
            if (det_value == 0 or is_new_file or is_end_of_vector) and within_det: #end current detection
                if is_end_of_vector:
                    end = det_index
                else:
                    end = prev_det_index
                
                within_det = False
                filename = filename_vector[start]

                # From all timestamps within the batch, select only the timestamps for the file containing the detection start
                file_timestamps = batch_support_data[filename_vector==filename]

                batch_start_timestamp = float(file_timestamps[0,1])
                batch_end_timestamp = float(file_timestamps[-1,1])
                time_start, duration = map_detection_to_time(start, end, step=step, spec_dur=spec_dur, batch_start_timestamp=batch_start_timestamp, batch_end_timestamp=batch_end_timestamp, buffer=buffer)
                score = np.average(scores_vector[start:(end + 1)])
                if np.isnan(score) and end==start:
                    score = scores_vector[start]
                
                det_timestamps[i].append((filename, time_start, duration, score))

            prev_det_index = det_index #update index of previous instance

    if not isinstance(threshold, list): 
        det_timestamps = det_timestamps[0]

    return det_timestamps

def process_batch(batch_data, batch_support_data, model, buffer=0, step=0.5, spec_dur=3.0, threshold=0.5, win_len=1, group=False):
    """ Runs one batch of (overlapping) spectrogram throught the classifier.

        Args:
            batch_data: numpy array
                An array with shape n,f,t,  where n is the number of spectrograms in the batch, t is the number of time bins and f the number of frequency bins.
            batch_support_data: numpy array
                An array of shape n x 2, where n is the batch size. The second dimension contains the filename and the start timestamp for each input in the batch
            model: ketos model
                The ketos trained classifier
            buffer: float
                Time (in seconds) to be added around the detection
            step: float
                The time interval(in seconds) between the starts of each contiguous input spectrogram.
                For example, a step=0.5 indicates that the first spectrogram starts at time 0.0s (from the beginning of the audio file), the second at 0.5s, etc.
            spec_dur: float
                The duration of each input spectrogram in seconds
            threshold: float or list of floats
                Minimum score value for a time step to be considered as a detection.
            win_len:int
                The windown length for the moving average. Must be an odd integer. The default value is 5.
            group:bool
                If False, return the filename, start, duration and scores for each spectrogram with score above the threshold. In this case, the duration will always be the duration of a single spectrogram.
                If True (default), average scores over(overlapping) spectrograms and group detections that are immediatelly next to each other. In this case, the score given for that detection will be the
                average score of all spectrograms comprising the detection event.

        Returns:
            batch_detections: list
                An array with all the detections in the batch. Each detection (first dimension) consists of the filename, start, duration and score.
                The start is given in seconds from the beginning of the file and the duration in seconds. 
                If a list of threshold values is specified, the returned object will be a list of arrays with len(batch_detections) = len(thresholds)         

    """
    thresholds = threshold if isinstance(threshold, list) else [threshold]

    scores = model.run_on_batch(batch_data, return_raw_output=True)

    if win_len == 1:
        scores = scores[:,1]
    else:
        scores = compute_avg_score(scores[:,1], win_len=win_len)
        
    if group == True:
        batch_detections = group_detections(scores, batch_support_data, buffer=buffer, step=step, spec_dur=spec_dur, threshold=thresholds) 

    else:
        batch_detections = []

        for thres in thresholds:
            threshold_indices = scores >= thres
            batch_det = np.vstack([batch_support_data[threshold_indices,0], batch_support_data[threshold_indices,1], np.repeat(spec_dur, sum(threshold_indices)), scores[threshold_indices]])
            if batch_det.shape[1] == 0:
                batch_det = []
            else:
                batch_det = [(d[0], float(d[1]), float(d[2]), float(d[3])) for d in batch_det.T]

            batch_detections.append(batch_det)


    if not isinstance(threshold, list): 
        batch_detections = batch_detections[0]

    return batch_detections


def process(provider, **kwargs):
    """ Use classifier to process audio clips (waveform or spectrogram).

        Delegates to :func:`detection.process_audio_loader` or :func:`detection.process_batch_generator`
        depending on the type of the `provider` argument.

        See these functions for a description of the required input arguments.

        Args:
            provider: instance of ketos.audio.audio_loader.AudioFrameLoader or ketos.data_handling.data_feeding.BatchGenerator
    """
    if isinstance(provider, AudioLoader):
        return process_audio_loader(provider, **kwargs)

    elif isinstance(provider, BatchGenerator):
        return process_batch_generator(provider, **kwargs)


def process_audio_loader(audio_loader, model, batch_size=128, threshold=0.5, buffer=0, win_len=1, group=False, progress_bar=False, merge=False):
    """ Use an audio_loader object to compute spectrogram from the audio files and process them with the trained classifier.

        Args:
            audio_loader: a ketos.audio.audio_loader.AudioFrameLoader object
                An audio loader that computes spectrograms from the audio audio files as requested
            model: ketos model
                The ketos trained classifier
            batch_size:int
                The number of spectrogram to process at a time.
            threshold: float or list of floats
                Minimum score value for a time step to be considered as a detection.
            buffer: float
                Time (in seconds) to be added around the detection
            win_len:int
                The windown length for the moving average. Must be an odd integer. The default value is 5.   
            group:bool
                If False, return the filename, start, duration and scores for each spectrogram with score above the threshold. In this case, the duration will always be the duration of a single spectrogram.
                If True (default), average scores over(overlapping) spectrograms and group detections that are immediatelly next to each other. In this case, the score given for that detection will be the
                average score of all spectrograms comprising the detection event.
            progress_bar: bool
                Show progress bar.  
            merge: bool
                Apply :func:`merge_overlapping_detections` to the detections before they are returned. Default is False.

        Returns:
            detections: list
                List of detections.
                If a list of threshold values is specified, the returned object will be a list of lists with len(detections) = len(thresholds)         
    """
    assert isinstance(win_len, int) and win_len%2 == 1, 'win_len must be an odd integer'

    thresholds = threshold if isinstance(threshold, list) else [threshold]
        
    n_extend = int((win_len - 1) / 2)

    n_batches = audio_loader.num() // batch_size
    last_batch_size = batch_size + (audio_loader.num() % batch_size)

    if n_batches == 0: 
        batch_sizes = [audio_loader.num()]
    elif n_batches == 1:
        batch_sizes = [last_batch_size]
    else:
        batch_sizes = [batch_size + n_extend] + [batch_size + 2 * n_extend for _ in range(n_batches - 2)] + [last_batch_size + n_extend]

    detections = [[] for _ in range(len(thresholds))]
    specs_prev_batch = []
    duration = None
    step = 0

    for siz in tqdm(batch_sizes, disable = not progress_bar): 
        batch_data, batch_support_data = [], []

        # first, collect data from the last specs from previous batch, if any            
        for spec in specs_prev_batch: 
            batch_data.append(spec.data)
            support_data = (spec.filename, spec.offset)
            batch_support_data.append(support_data)
            duration = spec.duration()

        # then, load specs from present batch
        specs_prev_batch = []
        while len(batch_data) < siz:
            spec = next(audio_loader)
            batch_data.append(spec.data)
            support_data = (spec.filename, spec.offset)
            batch_support_data.append(support_data)
            if siz - len(batch_data) < 2 * n_extend: specs_prev_batch.append(spec) # store last few specs
            duration = spec.duration()

        if len(batch_support_data) >= 2:
            step = batch_support_data[1][1] - batch_support_data[0][1]

        if step <= 0: step = duration

        batch_support_data = np.array(batch_support_data)
        batch_data = np.array(batch_data)

        batch_detections = process_batch(batch_data=batch_data, batch_support_data=batch_support_data, model=model, threshold=thresholds, 
                                        buffer=buffer, step=step, spec_dur=duration, win_len=win_len, group=group)

        for i in range(len(thresholds)):
            if len(batch_detections[i]) > 0: detections[i] += batch_detections[i]


    if merge:
        for i in range(len(thresholds)):
            detections[i] = merge_overlapping_detections(detections[i])

    if not isinstance(threshold, list): 
        detections = detections[0]

    return detections


def process_batch_generator(batch_generator, model, duration=3.0, step=0.5, threshold=0.5, buffer=0, win_len=1, group=False, merge=False):
    """ Use a batch_generator object to process pre-computed spectrograms stored in an HDF5 database with the trained classifier.

        Args:
            batch_generator: a ketos.data_handling.data_feeding.BatchGenerator object
                A batch_generator that loads pre-computed spectrograms from the a HDF5 database files as requested
            model: ketos model
                The ketos trained classifier
            duration: float
                The duration of each input spectrogram in seconds
            step: float
                The time interval(in seconds) between the starts of each contiguous input spectrogram.
                For example, a step=0.5 indicates that the first spectrogram starts at time 0.0s (from the beginning of the audio file), the second at 0.5s, etc.
            threshold: float or list of floats
                Minimum score value for a time step to be considered as a detection.
            buffer: float
                Time (in seconds) to be added around the detection
            win_len:int
                The windown length for the moving average. Must be an odd integer. The default value is 5.      
            group:bool
                If False, return the filename, start, duration and scores for each spectrogram with score above the threshold. In this case, the duration will always be the duration of a single spectrogram.
                If True (default), average scores over(overlapping) spectrograms and group detections that are immediatelly next to each other. In this case, the score given for that detection will be the
                average score of all spectrograms comprising the detection event.
            merge: bool
                Apply :func:`merge_overlapping_detections` to the detections before they are returned. Default is False.

        Returns:
            detections: list
                List of detections.
                If a list of threshold values is specified, the returned object will be a list of lists with len(detections) = len(thresholds)         
    """ 
    thresholds = threshold if isinstance(threshold, list) else [threshold]

    detections = [[] for _ in range(len(thresholds))]

    for b in range(batch_generator.n_batches):
        batch_data, batch_support_data = next(batch_generator)

        batch_detections = process_batch(batch_data=batch_data, batch_support_data=batch_support_data, model=model, threshold=thresholds, 
                                        buffer=buffer, spec_dur=duration, step=step, win_len=win_len, group=group)
        
        for i in range(len(thresholds)):
            if len(batch_detections[i]) > 0: detections[i] += batch_detections[i]


    if merge:
        for i in range(len(thresholds)):
            detections[i] = merge_overlapping_detections(detections[i])

    if not isinstance(threshold, list): 
        detections = detections[0]

    return detections


def transform_batch(x,y):
    """ Transform the data loaded from the database to the format expected by process_batch

        Args:
            x: numpy array
                A batch of spectrograms of shape batch_size, time bins, frequency bins
            y: numpy array
                Suppporting information for the batch (the filename and offset fields from the hdf5 dataset)
        
        Returns:
            tuple:
                transformed_x:numpy array
                    x, unmodified
                transformed_y:numpy array
                    y reshaped to shape batch_size, 2. type converted to str (from bytes string).
         """
  
    transformed_x = x
    filenames = y['filename']
    timestamps = y['offset']
    transformed_y = np.column_stack((filenames, timestamps))
    transformed_y = transformed_y.astype(str)
        
    return transformed_x, transformed_y


def merge_overlapping_detections(detections):
    """ Merge overlapping or adjacent detections.

        The score of the merged detection is computed as the average of the individual detection scores.

        Note: The detections are assumed to be sorted by start time in chronological order.

        Args:
            detections: numpy.array
                List of detections
        
        Returns:
            merged: numpy.array
                List of merged detections
    """
    num_det = len(detections)
    if num_det <= 1: return detections

    merged = [detections[0]]
    for i in range(1, num_det):
        filename = detections[i][0]
        start = detections[i][1]
        duration = detections[i][2]
        end = start + duration
        score = detections[i][3]
        filename_prev = merged[-1][0]
        start_prev = merged[-1][1]
        duration_prev = merged[-1][2]
        end_prev = start_prev + duration_prev        
        score_prev = merged[-1][3]
        
        # detections do not overlap, nor are they adjacent
        if start > end_prev or filename != filename_prev:
            merged.append(detections[i])

        # detections overlap, or adjacent to one another
        else:
            merged_dur = max(end, end_prev) - min(start,start_prev)
            avg_score = 0.5 * (score + score_prev) 
            merged_det = (filename, start_prev, merged_dur, avg_score)
            merged[-1] = merged_det #replace 

    return merged
    
def save_detections(detections, save_to):
    """ Save the detections to a csv file

        The resulting .csv is separated by commas, with each row representing one detection and has the following columns:

            filename: The name of the audio file where the detection was registered. In case the detection starts in one file and ends in the next, \
                      the name of the first file is registered.

            start:    Start of the detection (in seconds from the beginning of the file)

            duration: Length of the detection (in seconds)

            score:    The sore given to the detection by the trained neural network ([0-1])

        Args:
            detections: numpy.array
                List of detections
            save_to:string
                The path to the .csv file where the detections will be saved.
                Example: "/home/user/detections.csv"
    """
    if len(detections) == 0: return

    a = np.array(detections)
    df = pd.DataFrame({'filename':a[:,0], 'start':a[:,1], 'duration':a[:,2], 'score':a[:,3]})
    include_header = not os.path.exists(save_to)
    df.to_csv(save_to, mode='a', index=False, header=include_header)





import pytest
from unittest.mock import Mock, MagicMock
from ketos.neural_networks.dev_utils.detection import *
from ketos.neural_networks.cnn import CNNInterface
from ketos.data_handling.data_feeding import BatchGenerator
import numpy as np
import pandas as pd
import os
import shutil
import tables
import json


current_dir = os.path.dirname(os.path.realpath(__file__))
path_to_assets = os.path.join(os.path.dirname(current_dir),"assets")
path_to_tmp = os.path.join(path_to_assets,'tmp')

@pytest.fixture
def batch():
    data = np.vstack([np.zeros((10,8,8)), np.ones((3,8,8)),np.zeros((10,8,8)), np.ones((3,8,8)),np.zeros((4,8,8))])
    support = np.array([('file_1.wav', i*0.5) for i in range(30)],dtype=[('filename', '|S10'), ('offset', '>f4')])
    return data, support

@pytest.fixture
def mock_audio_loader(batch):
    data, support = batch
    audio_loader = MagicMock()
    audio_loader.current = 0

    def mock_next(self):
        if self.current == 30: 
            self.current = 0
        spec = Mock()
        spec.data = data[self.current]
        spec.filename = support[self.current][0]
        spec.filename = spec.filename.decode()
        spec.offset = support[self.current][1]
        spec.duration = lambda: 3.0

        self.current += 1

        return spec
    
    def mock_num():
        return 30

    audio_loader.num = mock_num
    audio_loader.__next__ = mock_next
    #AudioFrameLoader.__next__.return_value = (data[batch_1[0]:batch_1[1]], support[batch_1[0]:batch_1[1]]) 

    return audio_loader



@pytest.fixture
def scores_and_support_1():
    scores = np.array([0,0,0,0.1,0.3,0.4,0.52,0.89,0.78,0.6,0.4,0.3,0.4,0.2,0.1,0,0,0,0,0,0.7,0.4,0.5,0.8,0.7,0.4,0.3,0,0,0])
    support = np.array([('file_1.wav', 0),
                         ('file_1.wav', 0.5),
                         ('file_1.wav', 1.0),
                         ('file_1.wav', 1.5),
                         ('file_1.wav', 2.0),
                         ('file_1.wav', 2.5),
                         ('file_1.wav', 3.0),
                         ('file_1.wav', 3.5),
                         ('file_1.wav', 4.0),
                         ('file_1.wav', 4.5),
                         ('file_1.wav', 5.0),
                         ('file_1.wav', 5.5),
                         ('file_1.wav', 6.0),
                         ('file_1.wav', 6.5),
                         ('file_1.wav', 7.0),
                         ('file_1.wav', 7.5),
                         ('file_1.wav', 8.0),
                         ('file_1.wav', 8.5),
                         ('file_1.wav', 9.0),
                         ('file_1.wav', 9.5),
                         ('file_1.wav', 10),
                         ('file_1.wav', 10.5),
                         ('file_1.wav', 11.0),
                         ('file_1.wav', 11.5),
                         ('file_1.wav', 12.0),
                         ('file_1.wav', 12.5),
                         ('file_1.wav', 13.0),
                         ('file_1.wav', 13.5),
                         ('file_1.wav', 14.0),
                         ('file_1.wav', 14.5),
   
    ])
    return scores, support


@pytest.mark.parametrize("score_vector, win_len,  expected", [(np.array([1.,1.,1.,1.,1.,1.]), 3, np.array([0.,1.,1.,1.,1.,0.])),
                                                                (np.array([0.,0.,0.,0.,0.,0.]), 3, np.array([0.,0.,0.,0.,0.,0.])),
                                                                (np.array([1.,1.,1.,1.,1.,1.]), 5, np.array([0.,0.,1.,1.,0.,0.])),
                                                                (np.array([0.,0.,0.,0.,0.,0.]), 3, np.array([0.,0.,0.,0.,0.,0.])),
                                                                (np.array([0.1,0.35,0.7,0.8,0.5,0.3,0.2,0,0.89,0.5]), 3, np.array([0,0.3833,0.6166,0.6666,0.5333,0.3333,0.1666,0.3633,0.4633,0])),
                                                                 
                                                                 ])
def test_compute_avg_score(score_vector, win_len, expected):
    avg = compute_avg_score(score_vector, win_len)
    assert len(avg) == len(score_vector)
    assert np.allclose(avg, expected,  rtol=1e-03)


def test_compute_avg_score_raise_exception():
    with pytest.raises(AssertionError):
        avg = compute_avg_score(np.array([1,1,1,1,1]), win_len=2)


@pytest.mark.parametrize("det_start, det_end, batch_start_timestamp, batch_end_timestamp, step, spec_dur, buffer, expected", 
                        [(0,1,300,600,0.5,3.0, 1.0, (300,2.0)),
                         (0,2,300,600,0.5,3.0, 1.0, (300,2.5)),
                         (0,3,300,600,0.5,3.0, 1.0, (300,3.0)),
                         (1,4,300,600,0.5,3.0, 1.0, (300,3.5)),
                         (2,5,300,600,0.5,3.0, 1.0, (300,4.0)),
                         (2,6,300,600,0.5,3.0, 1.0, (300,4.5)),
                         (3,7,300,600,0.5,3.0, 1.0, (300.5,4.5)),
                         (4,8,300,600,0.5,3.0, 1.0, (301,4.5)),
                         (1,2,300,600,0.5,3.0, 1.0, (300,2.5)),
                         (2,2,300,600,0.5,3.0, 1.0, (300.0,2.5)),
                         (20,25,300,600,0.5,3.0, 1.0, (309,5.0)),
                         ])

def test_map_detection_to_time(det_start, det_end, batch_start_timestamp, batch_end_timestamp, step, spec_dur, buffer, expected):
    mapped_detection = map_detection_to_time(det_start, det_end, batch_start_timestamp, batch_end_timestamp, step, spec_dur, buffer)
    assert mapped_detection == expected

def test_map_detection_to_time_det_end_exception():
    with pytest.raises(ValueError):
        map_detection_to_time(det_start=10, det_end=9, batch_start_timestamp=300, batch_end_timestamp=600, step=0.5, spec_dur=3.0, buffer=1)



@pytest.mark.parametrize("buffer, step, spec_dur, threshold, expected", 
                        [(1.0, 0.5, 3.0, 0.5, [('file_1.wav',  2.0, 4.0, 0.697),
                                               ('file_1.wav',  9.0, 2.5, 0.7),
                                               ('file_1.wav', 10.0, 3.5, 0.667),]),
                         (0.0, 0.5, 3.0, 0.5, [('file_1.wav',  3.0, 2.0, 0.697),
                                               ('file_1.wav', 10.0, 0.5, 0.7),
                                               ('file_1.wav', 11.0, 1.5, 0.667),]),
                         (0.0, 0.5, 3.0, 0.7, [('file_1.wav',  3.5, 1.0, 0.835),
                                               ('file_1.wav', 10.0, 0.5, 0.7),
                                               ('file_1.wav', 11.5, 1.0, 0.75),])
                        ])


def test_group_detections(scores_and_support_1, buffer, step, spec_dur, threshold, expected):
    scores, support = scores_and_support_1
    grp_det = group_detections(scores_vector=scores, batch_support_data=support, buffer=buffer, step=step, spec_dur=spec_dur, threshold=threshold)
    for i,d in enumerate(grp_det):
        assert d[0] == expected[i][0]
        assert d[1] == expected[i][1]
        assert d[2] == expected[i][2]
        assert np.isclose(d[3], expected[i][3], rtol=1e-03)
    

def test_process_batch_without_avg_and_group(batch):
    data, support = batch
    transformed_data, transformed_support = transform_batch(data,support)
    model = CNNInterface.load_model_file(os.path.join(path_to_assets, "test_model.kt"), os.path.join(path_to_tmp, "tmp_model"))

    expected_detections = [('file_1.wav', 5.0, 3.0, 1.0), ('file_1.wav', 5.5, 3.0, 1.0), ('file_1.wav', 6.0, 3.0, 1.0), ('file_1.wav', 11.5, 3.0, 1.0), ('file_1.wav', 12.0, 3.0, 1.0), ('file_1.wav', 12.5, 3.0, 1.0)]
    detections = process_batch(batch_data=transformed_data, batch_support_data=transformed_support, model=model, buffer=1.0, step=0.5, threshold=0.5, win_len=1, group=False)

    assert detections == expected_detections


def test_process_batch_without_avg_and_group_multiple_thresholds(batch):
    data, support = batch
    transformed_data, transformed_support = transform_batch(data,support)
    model = CNNInterface.load_model_file(os.path.join(path_to_assets, "test_model.kt"), os.path.join(path_to_tmp, "tmp_model"))

    expected_detections = [('file_1.wav', 5.0, 3.0, 1.0), ('file_1.wav', 5.5, 3.0, 1.0), ('file_1.wav', 6.0, 3.0, 1.0), ('file_1.wav', 11.5, 3.0, 1.0), ('file_1.wav', 12.0, 3.0, 1.0), ('file_1.wav', 12.5, 3.0, 1.0)]
    detections = process_batch(batch_data=transformed_data, batch_support_data=transformed_support, model=model, buffer=1.0, step=0.5, threshold=[0.5,0.75,1.1], win_len=1, group=False)

    assert detections[0] == expected_detections
    assert detections[1] == expected_detections
    assert len(detections[2]) == 0


def test_process_batch_with_avg_and_group(batch):
    data, support = batch
    transformed_data, transformed_support = transform_batch(data,support)
    model = CNNInterface.load_model_file(os.path.join(path_to_assets, "test_model.kt"), os.path.join(path_to_tmp, "tmp_model"))

    expected_detections = [('file_1.wav', 4.0, 3.5, 0.6), ('file_1.wav', 10.5, 3.5, 0.6)]
    detections = process_batch(batch_data=transformed_data, batch_support_data=transformed_support, model=model, buffer=1.0, step=0.5, threshold=0.5, win_len=5, group=True)

    assert detections == expected_detections


def test_process_audio_loader_with_avg_and_group(mock_audio_loader):
    model = CNNInterface.load_model_file(os.path.join(path_to_assets, "test_model.kt"), os.path.join(path_to_tmp, "tmp_model"))

    expected_detections = [('file_1.wav', 4.0, 3.5, 0.6), ('file_1.wav', 10.5, 3.5, 0.6)]
    detections = process_audio_loader(audio_loader=mock_audio_loader, batch_size=15, model=model, buffer=1.0, threshold=0.5, win_len=5, group=True)

    assert detections == expected_detections


def test_process_audio_loader_without_avg_and_group(mock_audio_loader):
    model = CNNInterface.load_model_file(os.path.join(path_to_assets, "test_model.kt"), os.path.join(path_to_tmp, "tmp_model"))
    expected_detections = [('file_1.wav', 5.0, 3.0, 1.0), ('file_1.wav', 5.5, 3.0, 1.0), ('file_1.wav', 6.0, 3.0, 1.0), ('file_1.wav', 11.5, 3.0, 1.0), ('file_1.wav', 12.0, 3.0, 1.0), ('file_1.wav', 12.5, 3.0, 1.0)]
    detections = process_audio_loader(audio_loader=mock_audio_loader, batch_size=15, model=model, buffer=1.0, threshold=0.5, win_len=1, group=False)

    assert detections == expected_detections


def test_process_audio_loader_with_group(mock_audio_loader):
    model = CNNInterface.load_model_file(os.path.join(path_to_assets, "test_model.kt"), os.path.join(path_to_tmp, "tmp_model"))

    expected_detections = [('file_1.wav',  5.0, 1.5, 1.0), ('file_1.wav', 11.5, 1.5, 1.0)]
    detections = process_audio_loader(audio_loader=mock_audio_loader, batch_size=15, model=model, buffer=0, threshold=0.5, win_len=1, group=True)

    assert detections == expected_detections


def test_process_batch_generator_with_avg_and_group(batch):
    data, support = batch
    model = CNNInterface.load_model_file(os.path.join(path_to_assets, "test_model.kt"), os.path.join(path_to_tmp, "tmp_model"))

    batch_generator =  BatchGenerator(batch_size=15, x=data, y=support,output_transform_func=transform_batch, shuffle=False)
    
    expected_detections = [('file_1.wav', 4.0, 3.5, 0.6), ('file_1.wav', 10.5, 3.5, 0.6)]
    detections = process_batch_generator(batch_generator=batch_generator, model=model, buffer=1.0, step=0.5, threshold=0.5, win_len=5, group=True)

    assert detections == expected_detections

def test_process_batch_generator_without_avg_and_group(batch):
    data, support = batch
    model = CNNInterface.load_model_file(os.path.join(path_to_assets, "test_model.kt"), os.path.join(path_to_tmp, "tmp_model"))

    batch_generator =  BatchGenerator(batch_size=3, x=data, y=support,output_transform_func=transform_batch, shuffle=False)
    
    expected_detections = [('file_1.wav', 5.0, 3.0, 1.0), ('file_1.wav', 5.5, 3.0, 1.0), ('file_1.wav', 6.0, 3.0, 1.0), ('file_1.wav', 11.5, 3.0, 1.0), ('file_1.wav', 12.0, 3.0, 1.0), ('file_1.wav', 12.5, 3.0, 1.0)]
    detections = process_batch_generator(batch_generator=batch_generator, model=model, buffer=1.0, step=0.5, threshold=0.5, win_len=1, group=False)

    assert detections == expected_detections


def test_transform_batch(batch):
    data, support = batch
    transformed_data, transformed_support = transform_batch(data,support)

    expected_data = np.vstack([np.zeros((10,8,8)), np.ones((3,8,8)),np.zeros((10,8,8)), np.ones((3,8,8)),np.zeros((4,8,8))])
    expected_support = np.array([('file_1.wav', str(i*0.5)) for i in range(30)],dtype='<U32')

    assert transformed_data.shape == (30,8,8)
    assert transformed_support.shape == (30,2)
    assert np.array_equal(transformed_data, expected_data)
    assert np.array_equal(transformed_support, expected_support)


@pytest.mark.parametrize("detections, expected", [(

                                                    #Parameter set 1
                                                    #input
                                                    [('file_1.wav', 1.0, 3.0, 1.0),
                                                     ('file_1.wav', 2.0, 0.5, 0.5),
                                                     ('file_1.wav', 10.0, 5, 0.8)],
                                                     
                                                     #expected output
                                                     [('file_1.wav', 1.0, 3.0, 0.75),
                                                     ('file_1.wav', 10.0, 5, 0.8),]),

                                                    
                                                    #Parameter set 2
                                                    #input
                                                    ([('file_1.wav', 1.0, 3.0, 1.0),
                                                     ('file_1.wav', 2.0, 4.0, 0.5),
                                                     ('file_1.wav', 10.0, 5, 0.8)],
                                                     
                                                     #expected output
                                                     [('file_1.wav', 1.0, 5.0, 0.75),
                                                     ('file_1.wav', 10.0, 5, 0.8),]),
                                               

                                                    #Parameter set 3
                                                    #input
                                                    ([('file_1.wav', 2.0, 4.0, 0.697),
                                                     ('file_1.wav', 9.0, 2.5, 0.7),
                                                     ('file_1.wav', 10.0, 3.5, 0.667)],
                                                     
                                                     #expected output
                                                     [('file_1.wav', 2.0, 4.0, 0.697),
                                                     ('file_1.wav', 9.0, 4.5, 0.6835),]),
                                                     
                                                     #Parameter set 4
                                                     #input
                                                     ([('file_1.wav', 2.0, 4.0, 0.8),
                                                     ('file_1.wav', 3.0, 2.5, 0.9),
                                                     ('file_1.wav', 10.0, 3.5, 0.667)],
                                                     
                                                     #expected output
                                                     [('file_1.wav', 2.0, 4.0, 0.85),
                                                     ('file_1.wav', 10.0, 3.5, 0.667)])   
                                                      
                                                      ])
def test_merge_overlapping_detections(detections, expected):
    merged_detections = merge_overlapping_detections(detections)
    for i,d in enumerate(merged_detections):
        assert d[0] == expected[i][0]
        assert d[1] == expected[i][1]
        assert d[2] == expected[i][2]
        assert np.isclose(d[3], expected[i][3], rtol=1e-03)
    

def test_save_detections():
    detections = [('file_1.wav', 2.0, 4.0, 0.85),
                  ('file_1.wav', 10.0, 3.5, 0.667)]
    
    expected_df = pd.DataFrame(detections, columns=['filename', 'start', 'duration', 'score'])  

    destination = os.path.join(path_to_tmp,"saved_detections.csv")
    if os.path.exists(destination): os.remove(destination)
    save_detections(detections, save_to=destination)

    assert os.path.exists(destination)
    det_df = pd.read_csv(destination)
    assert det_df.equals(expected_df)
    
    os.remove(destination)



def test_save_detections_add_to_existing_file():
    detections = [('file_1.wav', 2.0, 4.0, 0.85),
                  ('file_1.wav', 10.0, 3.5, 0.667)]
    
    destination = os.path.join(path_to_tmp,"saved_detections.csv")
    if os.path.exists(destination): os.remove(destination)
    #Save 3 times to the same file. Only the first should add headers
    save_detections(detections, save_to=destination) 
    save_detections(detections, save_to=destination)
    save_detections(detections, save_to=destination)

    expected_df = pd.DataFrame(detections*3, columns=['filename', 'start', 'duration', 'score'])  

    assert os.path.exists(destination)
    det_df = pd.read_csv(destination)
    assert det_df.equals(expected_df)
    
    os.remove(destination)




# ================================================================================ #
#   Authors: Fabio Frazao and Oliver Kirsebom                                      #
#   Contact: fsfrazao@dal.ca, oliver.kirsebom@dal.ca                               #
#   Organization: MERIDIAN (https://meridian.cs.dal.ca/)                           #
#   Team: Data Analytics                                                           #
#   Project: ketos                                                                 #
#   Project goal: The ketos library provides functionalities for handling          #
#   and processing acoustic data and applying deep neural networks to sound        #
#   detection and classification tasks.                                            #
#                                                                                  #
#   License: GNU GPLv3                                                             #
#                                                                                  #
#       This program is free software: you can redistribute it and/or modify       #
#       it under the terms of the GNU General Public License as published by       #
#       the Free Software Foundation, either version 3 of the License, or          #
#       (at your option) any later version.                                        #
#                                                                                  #
#       This program is distributed in the hope that it will be useful,            #
#       but WITHOUT ANY WARRANTY; without even the implied warranty of             #
#       MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the              #
#       GNU General Public License for more details.                               # 
#                                                                                  #
#       You should have received a copy of the GNU General Public License          #
#       along with this program.  If not, see <https://www.gnu.org/licenses/>.     #
# ================================================================================ #

""" inception sub-module within the ketos.neural_networks module

    This module provides classes that implement Inception Neural Networks.

    Contents:
        ConvBatchNormRelu class
        InceptionBlock class
        Inception class
        InceptionInterface

"""

import tensorflow as tf
from .dev_utils.nn_interface import RecipeCompat, NNInterface
import json




default_inception_recipe =  {'n_blocks':3,
                    'n_classes':2,
                    'initial_filters':16,        
                    'optimizer': RecipeCompat('Adam', tf.keras.optimizers.Adam, learning_rate=0.005),
                    'loss_function': RecipeCompat('BinaryCrossentropy', tf.keras.losses.BinaryCrossentropy),  
                    'metrics': [RecipeCompat('BinaryAccuracy',tf.keras.metrics.BinaryAccuracy),
                                RecipeCompat('Precision',tf.keras.metrics.Precision),
                                RecipeCompat('Recall',tf.keras.metrics.Recall)],
                    }

class ConvBatchNormRelu(tf.keras.Model):
    """ Convolutional layer with batch normalization and relu activation.

        Used in Inception  Blocks

        Args: 
            n_filters: int
                Number of filters in the convolutional layer
            filter_shape: int
                The filter (i.e.: kernel) shape. 
            strides: int
                Strides to be used for the convolution operation
            padding:str
                Type of padding: 'same' or 'valid'
        
    """


    def __init__(self, n_filters, filter_shape=3, strides=1, padding='same'):
        super(ConvBatchNormRelu, self).__init__()

        self.model = tf.keras.models.Sequential([
            tf.keras.layers.Conv2D(n_filters, filter_shape, strides=strides, padding=padding),
            tf.keras.layers.BatchNormalization(),
            tf.keras.layers.ReLU()
        ])

    def call(self, x, training=None):
        """Calls the model on new inputs.

        In this case call just reapplies all ops in the graph to the new inputs (e.g. build a new computational graph from the provided inputs).

        Args:
            inputs: Tensor or list of tensors
                A tensor or list of tensors
            
            training: Bool
                Boolean or boolean scalar tensor, indicating whether to run the Network in training mode or inference mode.

        Returns:
                A tensor if there is a single output, or a list of tensors if there are more than one outputs.
        """
        x = self.model(x, training=training)

        return x


class InceptionBlock(tf.keras.Model):
    """ Inception Block for the Inception Architecture

        Args:
            n_filters:int
               The number of filters (i.e.: channels) to be used in each convolutional layer of the block
            strides: int
                Strides used in the first 3 and and 5th convolutional layers of the block

    """

    def __init__(self, n_filters, strides=1):
        super(InceptionBlock, self).__init__()

        self.n_filters = n_filters
        self.strides = strides

        self.conv1 = ConvBatchNormRelu(self.n_filters, strides=self.strides)
        self.conv2 = ConvBatchNormRelu(self.n_filters, filter_shape=3, strides=self.strides)
        self.conv3_1 = ConvBatchNormRelu(self.n_filters, filter_shape=3, strides=self.strides)
        self.conv3_2 = ConvBatchNormRelu(self.n_filters, filter_shape=3, strides=1)

        self.pool = tf.keras.layers.MaxPooling2D(3, strides=1, padding='same')
        self.pool_conv = ConvBatchNormRelu(self.n_filters, strides=self.strides)

    def call(self, x, training=None):
        """Calls the model on new inputs.

        In this case call just reapplies all ops in the graph to the new inputs (e.g. build a new computational graph from the provided inputs).

        Args:
            inputs: Tensor or list of tensors
                A tensor or list of tensors
            
            training: Bool
                Boolean or boolean scalar tensor, indicating whether to run the Network in training mode or inference mode.

        Returns:
                A tensor if there is a single output, or a list of tensors if there are more than one outputs.
        """
        x1 = self.conv1(x, training=training)
        x2 = self.conv2(x, training=training)
        x3_1 = self.conv3_1(x, training=training)
        x3_2 = self.conv3_2(x3_1, training=training)
        x4 = self.pool(x)
        x4 = self.pool_conv(x4, training=True)

        out = tf.concat([x1, x2, x3_2, x4], axis=3)

        return out


class InceptionArch(tf.keras.Model):
    """ Implements an Inception network, building on InceptionBlocks

        Args:
            n_blocks:int
                Number of Inception Blocks
            n_classes:int
                Number of possible classes 
            initial_filters:int
                Number of filters (i.e.: channels) in the first block
            pre_trained_base: instance of InceptionArch
                A pre-trained inception model from which the residual blocks will be taken. 
                Use by the the clone_with_new_top method when creating a clone for transfer learning
   
    
    """

    def __init__(self, n_blocks, n_classes, pre_trained_base=None, initial_filters=16, **kwargs):
        super(InceptionArch, self).__init__(**kwargs)

        self.input_channels = initial_filters
        self.output_channels = initial_filters
        self.n_blocks = n_blocks
        self.n_classes = n_classes
        self.initial_filters = initial_filters

        if pre_trained_base:
            self.conv1 = pre_trained_base[0]
            self.blocks = pre_trained_base[1]
        else:

            self.conv1 = ConvBatchNormRelu(self.initial_filters)

            self.blocks = tf.keras.models.Sequential(name='dynamic-blocks')

            for block_id in range(self.n_blocks):
                for layer_id in range(2):

                    if layer_id == 0:
                        block = InceptionBlock(self.output_channels, strides=2)
                    else:
                        block = InceptionBlock(self.output_channels, strides=1)

                    self.blocks.add(block)

                self.output_channels *= 2

        self.avg_pool = tf.keras.layers.GlobalAveragePooling2D()
        self.dense = tf.keras.layers.Dense(self.n_classes)
        self.softmax = tf.keras.layers.Softmax()

    def call(self, inputs, training=None):
        """Calls the model on new inputs.

        In this case call just reapplies all ops in the graph to the new inputs (e.g. build a new computational graph from the provided inputs).

        Args:
            inputs: Tensor or list of tensors
                A tensor or list of tensors
            
            training: Bool
                Boolean or boolean scalar tensor, indicating whether to run the Network in training mode or inference mode.

        Returns:
                A tensor if there is a single output, or a list of tensors if there are more than one outputs.
        """
        output = self.conv1(inputs, training=training)
        output = self.blocks(output, training=training)
        output = self.avg_pool(output)
        output = self.dense(output)
        output = self.softmax(output)

        return output

    def freeze_init_layer(self):
        """Freeze the initial convolutional layer"""
        self.layers[0].trainable = False

    def unfreeze_init_layer(self):
        """Unfreeze the initial convolutional layer"""
        self.layers[0].trainable = True
    
    def freeze_block(self, block_ids):
        """ Freeze specific inception blocks

            Args:
                blocks_ids: list of ints
                    The block numbers to be freezed (starting from zero)
        """

        for block_id in block_ids:
            self.layers[1].layers[block_id].trainable = False

    def unfreeze_block(self, block_ids):
        """ Unfreeze specific inception blocks

            Args:
                blocks_ids: list of ints
                    The block numbers to be freezed (starting from zero)
        """
        for block_id in block_ids:
            self.layers[1].layers[block_id].trainable = True
    
    def freeze_top(self):
        """Freeze the classification block"""
        for layer in self.layers[2:]:
            layer.trainable = False
    
    def unfreeze_top(self):
        """Unfreeze the classification block"""
        for layer in self.layers[2:]:
            layer.trainable = True


    def get_feature_extraction_base(self):
        """ Retrive the feature extraction base (initial convolutional layer + residual blocks)
        
            Returns:
                list containing the feature extraction layers
        """

        return [self.conv1, self.blocks]

    def clone_with_new_top(self, n_classes=None, freeze_base=True):
        """ Clone this instance but replace the original classification top with a new (untrained) one
        
            Args:
                n_classes:int
                    The number of classes the new classification top should output.
                    If None(default), the original number of classes will be used.
                freeze_base:bool
                    If True, the weights of the feature extraction base will be froze (untrainable) in the new model.
                
            Returns:
                cloned_model: instance of InceptionArch
                    The new model with the old feature extraction base and new classification top.
         """
        if freeze_base == True:
            self.trainable = False

        if n_classes is None:
            n_classes = self.n_classes

        pre_trained_base = self.get_feature_extraction_base()
        cloned_model = type(self)(n_classes=n_classes, pre_trained_base=pre_trained_base)

        return cloned_model
    
        
class InceptionInterface(NNInterface):
    """ Creates an Inception model with the standardized Ketos interface.

        Args:
            num_blocks: int
                The number of inception blocks to be used. 
            
            n_classes:int
                The number of classes. The output layer uses a Softmax activation and
                will contain this number of nodes, resulting in model outputs with this
                many values summing to 1.0.

            initial_filters:int
                The number of filters used in the first ResNetBlock. Subsequent blocks 
                will have two times more filters than their previous block.

            optimizer: ketos.neural_networks.RecipeCompat object
                A recipe compatible optimizer (i.e.: wrapped by the ketos.neural_networksRecipeCompat class)

            loss_function: ketos.neural_networks.RecipeCompat object
                A recipe compatible loss_function (i.e.: wrapped by the ketos.neural_networksRecipeCompat class)

            metrics: list of ketos.neural_networks.RecipeCompat objects
                A list of recipe compatible metrics (i.e.: wrapped by the ketos.neural_networksRecipeCompat class).
                These metrics will be computed on each batch during training.
                
    """

    @classmethod
    def _build_from_recipe(cls, recipe, recipe_compat=True):
        """ Build an Inception model from a recipe.

            Args:
                recipe: dict
                    A recipe dictionary. optimizer, loss function
                    and metrics must be instances of ketos.neural_networks.RecipeCompat.
                    
                    Example recipe:
                    
                    >>> {{'n_blocks':3, # doctest: +SKIP
                    ...    'n_classes':2,
                    ...    'initial_filters':16,        
                    ...    'optimizer': RecipeCompat('Adam', tf.keras.optimizers.Adam, learning_rate=0.005),
                    ...    'loss_function': RecipeCompat('BinaryCrossentropy', tf.keras.losses.BinaryCrossentropy),  
                    ...    'metrics': [RecipeCompat('CategoricalAccuracy',tf.keras.metrics.CategoricalAccuracy)],
                    }


            Returns:
                An instance of InceptionInterface.

        """

        n_blocks = recipe['n_blocks']
        n_classes = recipe['n_classes']
        initial_filters = recipe['initial_filters']
        
        if recipe_compat == True:
            optimizer = recipe['optimizer']
            loss_function = recipe['loss_function']
            metrics = recipe['metrics']
            
        else:
            optimizer = cls._optimizer_from_recipe(recipe['optimizer'])
            loss_function = cls._loss_function_from_recipe(recipe['loss_function'])
            metrics = cls._metrics_from_recipe(recipe['metrics'])
            

        instance = cls(n_blocks=n_blocks, n_classes=n_classes, initial_filters=initial_filters, optimizer=optimizer, loss_function=loss_function, metrics=metrics)

        return instance

    @classmethod
    def _read_recipe_file(cls, json_file, return_recipe_compat=True):
        """ Read an Inception recipe saved in a .json file.

            Args:
                json_file:string
                    Full path (including filename and extension) to the .json file containing the recipe.
                return_recipe_compat:bool
                    If True, returns a dictionary where the optimizer, loss_function, metrics and 
                    secondary_metrics (if available) values are instances of the ketos.neural_networks.nn_interface.RecipeCompat.
                    The returned dictionary will be equivalent to:
                            
                            >>> {'n_blocks':3, # doctest: +SKIP
                            ... 'n_classes':2,
                            ... 'initial_filters':16,        
                            ... 'optimizer': RecipeCompat('Adam', tf.keras.optimizers.Adam, learning_rate=0.005),
                            ... 'loss_function': RecipeCompat('BinaryCrossentropy', tf.keras.losses.BinaryCrossentropy),  
                            ... 'metrics': [RecipeCompat('CategoricalAccuracy',tf.keras.metrics.CategoricalAccuracy)]}

                    If False, the optimizer, loss_function, metrics and secondary_metrics (if available) values will contain a
                    dictionary representation of such fields instead of the RecipeCompat objects:
                            >>> {'n_blocks':3, # doctest: +SKIP
                            ... 'n_classes':2,
                            ... 'initial_filters':16,        
                            ... 'optimizer': {'name':'Adam', 'parameters': {'learning_rate':0.005}},
                            ... 'loss_function': {'name':'BinaryCrossentropy', 'parameters':{}},  
                            ... 'metrics': [{'name':'CategoricalAccuracy', 'parameters':{}}]}

                Returns:
                    recipe, according to 'return_recipe_compat.

        """

        with open(json_file, 'r') as json_recipe:
            recipe_dict = json.load(json_recipe)

        optimizer = cls._optimizer_from_recipe(recipe_dict['optimizer'])
        loss_function = cls._loss_function_from_recipe(recipe_dict['loss_function'])
        metrics = cls._metrics_from_recipe(recipe_dict['metrics'])
        
        if return_recipe_compat == True:
            recipe_dict['optimizer'] = optimizer
            recipe_dict['loss_function'] = loss_function
            recipe_dict['metrics'] = metrics
            
        else:
            recipe_dict['optimizer'] = cls._optimizer_to_recipe(optimizer)
            recipe_dict['loss_function'] = cls._loss_function_to_recipe(loss_function)
            recipe_dict['metrics'] = cls._metrics_to_recipe(metrics)

        return recipe_dict

    def __init__(self, n_blocks=default_inception_recipe['n_blocks'], n_classes=default_inception_recipe['n_classes'], initial_filters=default_inception_recipe['initial_filters'],
                       optimizer=default_inception_recipe['optimizer'], loss_function=default_inception_recipe['loss_function'], metrics=default_inception_recipe['metrics']):
        super(InceptionInterface, self).__init__(optimizer, loss_function, metrics)
        self.n_blocks = n_blocks
        self.n_classes = n_classes
        self.initial_filters = initial_filters

        self.model=InceptionArch(n_blocks=n_blocks, n_classes=n_classes, initial_filters=initial_filters)

    def _extract_recipe_dict(self):
        """ Create a recipe dictionary from an InceptionInterface instance.

            The resulting recipe contains all the fields necessary to build the same network architecture used by the instance calling this method.
            
            Returns:
                recipe:dict
                    A dictionary containing the recipe fields necessary to build the same network architecture.
                    The output is equivalent to:
                        >>> {'n_blocks':3, # doctest: +SKIP
                        ...    'n_classes':2,
                        ...    'initial_filters':16,        
                        ...    'optimizer': RecipeCompat('Adam', tf.keras.optimizers.Adam, learning_rate=0.005),
                        ...    'loss_function': RecipeCompat('BinaryCrossentropy', tf.keras.losses.BinaryCrossentropy),  
                        ...    'metrics': [RecipeCompat('CategoricalAccuracy',tf.keras.metrics.CategoricalAccuracy)]}
        """

        recipe = {}
        recipe['interface'] = type(self).__name__
        recipe['n_blocks'] = self.n_blocks
        recipe['n_classes'] = self.n_classes
        recipe['initial_filters'] = self.initial_filters
        recipe['optimizer'] = self._optimizer_to_recipe(self.optimizer)
        recipe['loss_function'] = self._loss_function_to_recipe(self.loss_function)
        recipe['metrics'] = self._metrics_to_recipe(self.metrics)
        
        return recipe


